; Si vuole programmare un sistema a processore in grado di gestire
; linserimento e lestrazione di valori su 8 bit in/da un buffer
; circolare. Il sistema a processore esegue le seguenti operazioni:
;   - acquisisce valori su 8 bit (interpretati come numeri senza segno)
;     che giungono attraverso la porta A dell'8255 (programmata in modo
;     1 e connessa allinterrupt 39);
;   - tali valori vengono memorizzati nel buffer circolare: se il buffer
;      pieno, il dato in arrivo viene scartato
;   - se viene ricevuto il valore 0, non si inserisce nulla, si estrae
;     un elemento dal buffer e lo si invia tramite la porta B.
#start=8255.exe#

	LEN equ 5
     
.model small          
.data

	BUFFER db LEN dup(0)
	COUNT  dw 0
	TESTA  dw 0
	CODA   dw 0 
	
.stack     
.code   

			; procedura di inizializzazione della interrupt vector table
INIT_IVT	proc
			push ax
			push bx
			push dx
			push ds
			xor	ax, ax
			mov	ds, ax      
		
			; channel 7
			mov	bx, 39		
			shl	bx, 2		
			mov ax, offset ISR_PA_IN
			mov	ds:[bx], ax
			mov	ax, seg ISR_PA_IN
			mov	ds:[bx+2], ax       
			; channel 6
			mov	bx, 38		
			shl	bx, 2		
			mov ax, offset ISR_PA_OUT
			mov	ds:[bx], ax
			mov	ax, seg ISR_PA_OUT
			mov	ds:[bx+2], ax       		
			; channel 5
			mov	bx, 37		
			shl	bx, 2		
			mov ax, offset ISR_PB_IN
			mov	ds:[bx], ax
			mov	ax, seg ISR_PB_IN
			mov	ds:[bx+2], 	AX       				
			; channel 4
			mov	bx, 36		
			shl	bx, 2		
			mov ax, offset ISR_PB_OUT
			mov	ds:[bx], ax
			mov	ax, seg ISR_PB_OUT
			mov	ds:[bx+2], ax       		
					
			pop	ds
			pop	dx
			pop	bx
			pop	ax
			ret
INIT_IVT	endp

INIT_8255	proc
		    mov al, 10110000b
		    out 083h, al
		    mov al, 00001001b
		    out 083h, al
            ret
INIT_8255   endp        

; ISR for reading the value received on PA            
ISR_PA_IN   proc
            ; aquisizione valori da portA 8255              
            in al, 080h
            cmp al, '0'
            ; se e' zero dobbiamo eliminare dal buffer
            jz is_zero
            ; controlliamo se il buffer e' pieno
            cmp COUNT, LEN
            jz fine
            ; inseriamo in coda, la testa non cambia
            mov si, CODA
            mov BUFFER[si], al
            inc si
            cmp si, LEN
            ; se e' pari alla lunghezza la coda ritorna in posizione 0
            jz reset_si       
cont:       mov CODA, si
            mov ax, COUNT
            inc ax
            mov COUNT, ax
            jmp fine
is_zero:    ; controlliamo se il buffer e' vuoto
            cmp COUNT, 0
            jz  fine
            ; elimino dalla testa che avanza di una posizione
            mov si, TESTA
            mov al, BUFFER[si]
            mov BUFFER[si], byte ptr 0
            out 081h, al
            inc si
            cmp si, LEN
            jz reset_si_2
cont2:      mov TESTA, si
            mov ax, COUNT
            dec ax
            mov COUNT, ax
            cmp ax, 0
            ; se il buffer e' vuoto non ha senso continuare
            ; a scrivere dalla coda, ma dall'inizio
            jz reset_t_c
            jmp fine
reset_si:   xor si, si
            jmp cont
reset_si_2: xor si, si
            jmp cont2
reset_t_c:  mov TESTA, word ptr 0
            mov CODA, word ptr 0
fine:       sti
            iret    
ISR_PA_IN   endp             
              
; ISR for waiting a confirmation that the value written on PA is externally read 
ISR_PA_OUT  proc               
            iret    
ISR_PA_OUT  endp                                  
                           
; ISR for reading the value received on PB                                       
ISR_PB_IN   proc
            iret    
ISR_PB_IN   endp             
                                          
; ISR for waiting a confirmation that the value written on PB is externally read                                           
ISR_PB_OUT  proc
            iret    
ISR_PB_OUT  endp         
            
.startup
    
		cli         
		call INIT_IVT
		call INIT_8255
		sti            
              
next:  	nop         
		;hlt
		jmp next

.exit
end
