; Si programmi il sistema in modo che
;    a. Legga dati su 16 bit dalla porta A dell'8255 (viene inviato
;       prima LSB poi MSB). I dati devono essere letti non appena
;       disponibili.
;    b. Ogni secondo il sistema deve inviare sulla porta B dell8259
;       alternativamente
;           i. L'ultimo valore acquisito (16 bit prima MSB, poi LSB)
;           ii. Il valore 0.
;    c. Se riceve il valore 0FFFFH sulla porta A dell'8255, il sistema
;       si prepara per lo spegnimento (switch-off)
;           i. Programma un'attesa di 5 minuti, durante la quale sia
;              ancora possibile effettuare richieste di interrupt.
;           ii. Al termine dellattesa blocca le richieste di interrupt
;           iii. Infine per un ulteriore 1 minuto emette sulla porta B
;                il valore 0FFH (segnalazione di spegnimento entro 1
;                minuto), poi il valore 000H.
#start=8255.exe#
#start=8253.exe#
    .model small          
    .data
        
        VAL         dw 0
        WAITING     db 0
        SECS        dw 0
                
    .stack     
    .code   
    
    
    ; procedura di inizializzazione della interrupt vector table
INIT_IVT	PROC
		PUSH 	AX
		PUSH	BX
		PUSH	DX
		PUSH 	DS
		XOR	AX, 	AX
		MOV	DS, 	AX      
		
		; channel 7
		MOV	BX, 	39		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PA_IN
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PA_IN
		MOV	DS:[BX+2], 	AX       
		; channel 6
		MOV	BX, 	38		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PA_OUT
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PA_OUT
		MOV	DS:[BX+2], 	AX       		
		; channel 5
		MOV	BX, 	37		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PB_IN
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PB_IN
		MOV	DS:[BX+2], 	AX       				
		; channel 4
		MOV	BX, 	36		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PB_OUT
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PB_OUT
		MOV	DS:[BX+2], 	AX
		; channel 3
		mov bx, 35
		shl bx, 2
		mov ax, offset ISR_COUNTER0
		mov ds:[bx], ax
		mov ax, seg ISR_COUNTER0
		mov ds:[bx+2], ax
		; channel 2
		mov bx, 34
		shl bx, 2
		mov ax, offset ISR_COUNTER12
		mov ds:[bx], ax
		mov ax, seg ISR_COUNTER12
		mov ds:[bx+2], ax       		
					
		POP	DS
		POP	DX
		POP	BX
		POP	AX
		RET
INIT_IVT	ENDP

INIT_8255	PROC
			mov al, 10110000b
			out 083h, al
			mov al, 00001001b ; abilito interrupt
			out 083h, al
            RET
INIT_8255   ENDP

INIT_8253 proc
		  ; 3.000.000 di colpi = 3.000 COUNTER1 * 1.000 COUNTER2
		  mov al, 01110100b
		  out 063h, al
		  mov ax, 3000-1
		  out 061h, al
		  xchg al, ah
		  out 061h, al
		  
		  mov al, 10110100b
		  out 063h, al
		  mov ax, 1000-1
		  out 062h, al
		  xchg al, ah
		  out 062h, al
		  
          ret
INIT_8253 endp

; ISR for reading the value received on PA            
ISR_PA_IN   PROC
			in al, 080h
			xchg al, ah
			in al, 080h
			mov VAL, ax
            IRET    
ISR_PA_IN   ENDP             
              
; ISR for waiting a confirmation that the value written on PA is externally read 
ISR_PA_OUT  PROC               
            IRET    
ISR_PA_OUT  ENDP                                  
                           
; ISR for reading the value received on PB                                       
ISR_PB_IN   PROC              
			IRET    
ISR_PB_IN   ENDP             
                                          
; ISR for waiting a confirmation that the value written on PB is externally read                                           
ISR_PB_OUT  PROC
            IRET    
ISR_PB_OUT  ENDP 

ISR_COUNTER0  proc    
			  iret
ISR_COUNTER0  endp
                

ISR_COUNTER12 proc
			  cmp WAITING, byte ptr 1
			  jz  inc_secs_5_m
			  cmp WAITING, byte ptr 2
			  jz  inc_secs_1_m
			  cmp WAITING, byte ptr 3
			  jz  fine
              mov ax, VAL
			  cmp ax, 0ffffh
			  jz switch_off
cont:         mov ax, VAL
			  xchg al, ah
			  out 081h, al
			  xchg al, ah
			  out 081h, al
			  mov al, 0
			  out 081h, al
			  jmp fine

switch_off:   mov WAITING, byte ptr 1
              jmp fine

inc_secs_5_m: mov ax, SECS
              inc ax
              mov SECS, ax
              cmp SECS, 300
              jz last_wait
              jmp cont

last_wait:    mov WAITING, byte ptr 2
              mov SECS, word ptr 0
              jmp fine
              
inc_secs_1_m: mov al, 0ffh
              out 081h, al 
              mov ax, SECS
              inc ax
              mov SECS, ax
              cmp SECS, 60
              jz  final_off
              ; disabilito interrupt della porta A
              mov al, 00001000b
			  out 083h, al
              jmp fine
              
final_off:    mov WAITING, byte ptr 3
              mov al, 000h
              out 081h, al
			  
fine:    	  iret
ISR_COUNTER12 endp
            
    
.startup    
       CLI         
       call INIT_IVT
       call INIT_8255
       call INIT_8253
       STI            
next:  nop         
       ;hlt

       jmp next
          
.exit
end
