; Si consideri un sistema 8086 avente l'architettura riportata in figura.
; La porte A dell8255  configurata per funzionare in modo 1 in input,
; mentre la porta B  impostata in modo 0 output. Le porte servono per scambiare
; dati su 8 bit con l'esterno, supportate da una struttura di memorizzazione FIFO.
; Il sistema riceve dall'esterno sequenze di comandi (corrispondenti ciascuno ad un
; carattere inviato sulla porta A dell'8255 inferiore) e dati. I comandi possibili sono:
;    I: acquisisci dalla porta A dell'8255 un dato (intero senza segno su 8 bit)
;       e memorizzalo in una struttura FIFO interna
;    O: estrai dalla struttura FIFO un dato e invialo sulla porta B dell'8255
;    Z: azzera la struttura FIFO interna. A seguito del comando I sulla porta A,
;       il sistema riceve 1 o + dati che vengono inseriti nella FIFO (fino eventualmente
;       alla saturazione). A seguito del comando O, il sistema estrae un singolo dato e
;       lo emette sulla porta B. Si consideri inoltre:
;    - in caso siano ricevuti dati sulla porta A senza aver ricevuto il comando I
;      (ad esempio dopo aver ricevuto O o Z), tali dati devono essere ignorati
;    - in caso la struttura FIFO sia piena, alla ricezione di un nuovo dato,
;      il sistema deve emettere il valore 0FFH sulla porta B (indicazione di saturazione della FIFO)
; Il sistema deve infine emettere sulla porta B il numero di dati ricevuti ogni 10 ms;
; si consideri tale dato su 16 bit.

#start=8255.exe#
#start=8253.exe#

    LEN equ 5
    
    .model small          
    .data
            FIFO        db  LEN dup(?)
            I_RECEIVED  db  0
            N           dw  0
                  
    .stack     
    .code   
    
    
    ; procedura di inizializzazione della interrupt vector table
INIT_IVT	PROC
		PUSH 	AX
		PUSH	BX
		PUSH	DX
		PUSH 	DS
		XOR	AX, 	AX
		MOV	DS, 	AX      
		
		; channel 7
		MOV	BX, 	39		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PA_IN
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PA_IN
		MOV	DS:[BX+2], 	AX       
		; channel 6
		MOV	BX, 	38		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PA_OUT
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PA_OUT
		MOV	DS:[BX+2], 	AX       		
		; channel 5
		MOV	BX, 	37		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PB_IN
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PB_IN
		MOV	DS:[BX+2], 	AX       				
		; channel 4
		MOV	BX, 	36		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_PB_OUT
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_PB_OUT
		MOV	DS:[BX+2], 	AX     
        ; channel 3
		MOV	BX, 	35		
		SHL	BX, 	2		
		MOV AX, 	offset ISR_COUNT0
		MOV	DS:[BX], 	AX
		MOV	AX,     seg ISR_COUNT0
		MOV	DS:[BX+2], 	AX  		
					
		POP	DS
		POP	DX
		POP	BX
		POP	AX
		RET
INIT_IVT	ENDP

INIT_8255	PROC
		    mov al, 10110000b
		    out 083h, al
		    mov al, 00001001b    
		    out 083h, al   
            RET
INIT_8255   ENDP

INIT_8253	PROC
            mov al, 00110100b
            out 063h, al
            mov ax, 30000-1
            out 060h, al
            xchg al, ah
            out 060h, al
            RET
INIT_8253   ENDP

; ISR for reading the value received on PA            
ISR_PA_IN   PROC                
            in al, 080h       ;aquisizione valori da portA 8255  
            cmp al, 'I'
            jz is_I
            cmp al, 'O'
            jz is_O
            cmp al, 'Z'
            jz is_Z
            cmp I_RECEIVED, 1
            jnz fine
            cmp N, LEN
            jz is_full
store:      mov bx, N
            mov FIFO[bx], al
            inc bx
            mov N, bx
            jmp fine
            
           
is_I:       cmp I_RECEIVED, byte ptr 1
            jz  store
            mov I_RECEIVED, byte ptr 1
            jmp fine
            
is_O:       mov I_RECEIVED, byte ptr 0
            cmp N, 0
            jz fine
            mov al, FIFO[0]
            out 081h, al
            mov cx, N
            dec cx
            mov N, cx
            cmp N, word ptr 0
            jz fine
            mov si, 1
ciclo:      mov al, FIFO[si]
            mov FIFO[si-1], al
            inc si
            dec cx
            jnz ciclo
            mov FIFO[si-1], byte ptr 0            
            jmp fine
            
is_Z:       mov I_RECEIVED, byte ptr 0
            mov cx, N
            mov si, 0
ciclo2:     mov FIFO[si], byte ptr 0
            inc si
            dec cx
            jnz ciclo2
            mov N, word ptr 0
            jmp fine
is_full:    mov al, 0ffh
            out 081h, al
            
fine:       IRET    
ISR_PA_IN   ENDP             
              
; ISR for waiting a confirmation that the value written on PA is externally read 
ISR_PA_OUT  PROC               

            IRET    
ISR_PA_OUT  ENDP                                  
                           
; ISR for reading the value received on PB                                       
ISR_PB_IN   PROC              
            in      al, 081h                                    
            IRET    
ISR_PB_IN   ENDP             
                                          
; ISR for waiting a confirmation that the value written on PB is externally read                                           
ISR_PB_OUT  PROC


            IRET    
ISR_PB_OUT  ENDP

ISR_COUNT0  PROC
            mov ax, N
            out 081h, al
            xchg al, ah
            out 081h, al
            IRET    
ISR_COUNT0  ENDP 
            
    
       .startup    
       CLI         
       call INIT_IVT
       call INIT_8255
       call INIT_8253
       STI            
              
next:  nop         
      ;hlt

      jmp next
          
    
    
    .exit

    end
;pb