;*******************************************************************************************
;* Startup.s: Startup file for the uCdragor board based on LPC2292-ARM7T     			   *
;*******************************************************************************************
;* Author: Paolo Bernardi - Politecnico di Torino										   *	
;* Revision: r1 Francesco Brundu 176471
;*******************************************************************************************
;*******************************************************************************************
;* Functionalities: 														 			   *
;*	+ simulates traffic lights															   *
;*******************************************************************************************

;###########################################################################################
; constants 
;###########################################################################################

; Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs
Mode_USR        EQU     0x10
Mode_FIQ        EQU     0x11
Mode_IRQ        EQU     0x12
Mode_SVC        EQU     0x13
Mode_ABT        EQU     0x17
Mode_UND        EQU     0x1B
Mode_SYS        EQU     0x1F

I_Bit           EQU     0x80            ; when I bit is set, IRQ is disabled
F_Bit           EQU     0x40            ; when F bit is set, FIQ is disabled

;// <h> Stack Configuration (Stack Sizes in Bytes)

UND_Stack_Size  EQU     0x00000000
SVC_Stack_Size  EQU     0x00000008
ABT_Stack_Size  EQU     0x00000000
FIQ_Stack_Size  EQU     0x00000000
IRQ_Stack_Size  EQU     0x00000080
USR_Stack_Size  EQU     0x00000400	; 1 KByte

; calcola la somma della dimensione dello stack ad eccezione di quella della
; modalit user
ISR_Stack_Size  EQU     (UND_Stack_Size + SVC_Stack_Size + ABT_Stack_Size + FIQ_Stack_Size + IRQ_Stack_Size)
			    
; serve per dire al compilatore che ci serve un area stack, non inizializzata,
; in lettura e scrittura e con un allineamento di 4 byte
                AREA    STACK, NOINIT, READWRITE, ALIGN=3

; mnemonico al quale  associato uno spazio di USR_Stack_Size
Stack_Mem       SPACE   USR_Stack_Size
__initial_sp    SPACE   ISR_Stack_Size
Stack_Top

;// <h> Heap Configuration
Heap_Size       EQU     0x00000400	  ;1 KByte

                AREA    HEAP, NOINIT, READWRITE, ALIGN=3
__heap_base
Heap_Mem        SPACE   Heap_Size
__heap_limit

; VPBDIV definitions Peripheral Bus Clock Rate
VPBDIV          EQU     0xE01FC100      ; VPBDIV Address
VPBDIV_SETUP    EQU     0				; cpu clock/4
VPBDIV_Val      EQU     0x00000000

; Phase Locked Loop (PLL) definitions
PLL_BASE        EQU     0xE01FC080      ; PLL Base Address
PLLCON_OFS      EQU     0x00            ; PLL Control Offset
PLLCFG_OFS      EQU     0x04            ; PLL Configuration Offset
PLLSTAT_OFS     EQU     0x08            ; PLL Status Offset
PLLFEED_OFS     EQU     0x0C            ; PLL Feed Offset
PLLCON_PLLE     EQU     (1<<0)          ; PLL Enable
PLLCON_PLLC     EQU     (1<<1)          ; PLL Connect
PLLCFG_MSEL     EQU     (0x1F<<0)       ; PLL Multiplier
PLLCFG_PSEL     EQU     (0x03<<5)       ; PLL Divider
PLLSTAT_PLOCK   EQU     (1<<10)         ; PLL Lock Status

PLL_SETUP       EQU     1
PLLCFG_Val      EQU     0x00000024


; Memory Accelerator Module (MAM) definitions
MAM_BASE        EQU     0xE01FC000      ; MAM Base Address
MAMCR_OFS       EQU     0x00            ; MAM Control Offset
MAMTIM_OFS      EQU     0x04            ; MAM Timing Offset

MAM_SETUP       EQU     1
MAMCR_Val       EQU     0x00000002
MAMTIM_Val      EQU     0x00000004

; Memory remap 
MEMMAP          EQU     0xE01FC040

; External Memory Controller (EMC) definitions
EMC_BASE        EQU     0xFFE00000      ; EMC Base Address
BCFG0_OFS       EQU     0x00            ; BCFG0 Offset
BCFG1_OFS       EQU     0x04            ; BCFG1 Offset
BCFG2_OFS       EQU     0x08            ; BCFG2 Offset
BCFG3_OFS       EQU     0x0C            ; BCFG3 Offset

;External Memory Controller (EMC)
EMC_SETUP   	EQU     0
BCFG0_SETUP 	EQU     0
BCFG0_Val   	EQU     0x0000FBEF
BCFG1_SETUP 	EQU     0
BCFG1_Val   	EQU     0x0000FBEF
BCFG2_SETUP 	EQU     0
BCFG2_Val   	EQU     0x0000FBEF
BCFG3_SETUP 	EQU     0
BCFG3_Val   	EQU     0x0000FBEF

; Connect block setup
PINSEL0			EQU		0XE002C000		; PINSEL0 Address
PINSEL1			EQU		0XE002C004		; PINSEL1 Address
PINSEL2     	EQU     0xE002C014      ; PINSEL2 Address

PINSEL0_val		EQU		0x55550005 
	;//*** 32bit PORT 0 P0.0--P0.15 GROUP function configuration
	;//Ports  :  P0.15    P0.14    P0.13    P0.12	P0.11    P0.10    P0.09    P0.08
	;//Signal :  RI  	 CD 	  DTR     DSR      CTS  	RTS 	 RXD1     TXD1 
	;//Binary :  01  	 01 	  01	   01   	01  	 01 	  01	   01
	;//______________________________________________________________________________
	;//Ports  :  P0.07    P0.06    P0.05    P0.04	P0.03    P0.02    P0.01    P0.00
	;;//Signal :  GPIO	 GPIO     GPIO     GPIO 	GPIO	 GPIO     RXD0     TXD0 
	;//Binary :  00  	 00 	  00	   00   	00  	 00 	  01	   01
	;//******************************************************************************	

PINSEL1_val		EQU		0x25450400
	;//*** 32bit PORT 0 P0.16--P0.31 GROUP function configuration_____________gongjun
	;//Ports  :  P0.31    P0.30    P0.29    P0.28	P0.27    P0.26    P0.25    P0.24
	;//Signal :  GPIO	 GPIO     AIN2     AIN1 	AIN0	 GPIO     RD1      TD2 
	;//Binary :  00  	 01 	  01	   01   	01  	 00 	  01	   01
	;//______________________________________________________________________________
	;//Ports  :  P0.23    P0.22    P0.21    P0.20	P0.19    P0.18    P0.17    P0.16
	;//Signal :  RD2 	GPIO	 PWM5     GPIO     GPIO 	GPIO	 GPIO     GPIO 
	;//Binary :  01  	 00 	  01	   00   	00  	 00 	  00	   00
	;//******************************************************************************

PINSEL2_val 	EQU		0x0F814910
	;//*** 32bit PORT 1, PORT 2, PORT 3 GROUP function configuration
	;//Ports  :  RESERVED P3.22:2  P3.1
	;//Signal :  GPIO	 ADDR2:23 ADDR1 
	;//Binary :  00  	 111	   1
	;//______________________________________________________________________________
	;//Ports  :  P3.00    P2.31    P2.30    P2.29:28 RESERVED P3.24
	;//Signal :  ADDR0   GPIO	 GPIO     GPIO     GPIO 	CS3 
	;//Binary :  1   	0   	 0  	  0 	   00   	01
	;//______________________________________________________________________________
	;//Ports  :  P3.25    P3.23    RESERVER P3.26	RESERVER RESERVER P3.27
	;//Signal :  CS2 	 ADDR23   GPIO     CS1  	GPIO	 GPIO     nWE 
	;//Binary :  01  	 1  	  0 	   1		0   	 0  	  1
	;//______________________________________________________________________________
	;//Ports  :  P3.28    P3.29    P1.2.3   P1.25:16 P1.31:26 RESERVER RESERVER
	;//Signal :  GPIO	 GPIO     GPIO     DEBUG	GPIO	 GPIO     GPIO
	;//Binary :  0   	 0  	  01	   1		0   	 0  	  0
	;//******************************************************************************

; GPIO setup
;  port0
IO0PIN			EQU		0xE0028000
IO0SET			EQU		0xE0028004
IO0DIR			EQU		0xE0028008
IO0CLR			EQU		0xE002800C
;  port 1
IO1PIN			EQU		0xE0028010
IO1SET			EQU		0xE0028014
IO1DIR			EQU		0xE0028018
IO1CLR			EQU		0xE002801C
;  port 2
IO2PIN			EQU		0xE0028020
IO2SET			EQU		0xE0028024
IO2DIR			EQU		0xE0028028
IO2CLR			EQU		0xE002802C
;  port 3
IO3PIN			EQU		0xE0028030
IO3SET			EQU		0xE0028034
IO3DIR			EQU		0xE0028038
IO3CLR			EQU		0xE002803C

; semaforo automobili (led gruppo destra)
IO0DIR_val		EQU 	0x00280035
	;//*** 32bit PORT 0 P0.0--P0.31 GROUP direction configuration
	;//Ports  :  P0.31    P0.30    P0.29    P0.28	P0.27    P0.26    P0.25    P0.24
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P0.23    P0.22    P0.21    P0.20	P0.19    P0.18    P0.17    P0.16
	;//Signal :  I   	 I  	  o 	   I		o   	 I  	  I 	   I
	;//Binary :  0   	 0  	  1 	   0		1   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P0.15    P0.14    P0.13    P0.12	P0.11    P0.10    P0.09    P0.08
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P0.07    P0.06    P0.05    P0.04	P0.03    P0.02    P0.01    P0.00
	;//Signal :  I   	 I  	  o 	   o		I   	 o  	  I 	   o
	;//Binary :  0   	 0  	  1 	   1		0   	 1  	  0 	   1
	;//******************************************************************************
IO0_red			EQU		0x00000014
IO0_yel			EQU 	0x00000024
IO0_gre			EQU 	0x00000030

IO1DIR_val		EQU		0x01E60000
	;//*** 32bit PORT 0 P1.00--P1.31 GROUP direction configuration
	;//Ports  :  P1.31    P1.30    P1.29    P1.28	P1.27    P1.26    P1.25    P1.24
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   o
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   1
	;//______________________________________________________________________________
	;//Ports  :  P1.23    P1.22    P1.21    P1.20	P1.19    P1.18    P1.17    P1.16
	;//Signal :  o   	 o  	  o 	   I		I   	 o  	  o 	   I
	;//Binary :  1   	 1  	  1 	   0		0   	 1  	  1 	   0
	;//______________________________________________________________________________
	;//Ports  :  P1.15    P1.14    P1.13    P1.12	P1.11    P1.10    P1.09    P1.08
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P1.07    P1.06    P1.05    P1.04	P1.03    P1.02    P1.01    P1.00
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//******************************************************************************

; semaforo pedoni (gruppo led sinistra)
IO2DIR_val		EQU		0x78000000 	
	;//*** 32bit PORT 2 GROUP direction configuration
	;//Ports  :  P2.31    P2.30    P2.29    P2.28	P2.27    P2.26    P2.25    P2.24
	;//Signal :  I   	 o  	  o 	   o		o   	 I  	  I 	   I
	;//Binary :  0   	 1  	  1 	   1		1   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P2.23    P2.22    P2.21    P2.20	P2.19    P2.18    P2.17    P2.16
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P2.15    P2.14    P2.13    P2.12	P2.11    P2.10    P2.09    P2.08
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P2.07    P2.06    P2.05    P2.04	P2.03    P2.02    P2.01    P2.00
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//******************************************************************************
IO2_red			EQU		0x30000000 	
IO2_yel			EQU		0x50000000 	
IO2_gre			EQU		0x60000000 	
											
IO3DIR_val 		EQU		0x00000000
	;//*** 32bit PORT 3 GROUP direction configuration_____________gongjun
	;//Ports  :  P3.31    P3.30    P3.29    P3.28	P3.27    P3.26    P3.25    P3.24
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;;//______________________________________________________________________________
	;//Ports  :  P3.23    P3.22    P3.21    P3.20	P3.19    P3.18    P3.17    P3.16
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;;//Ports  :  P3.15    P3.14    P3.13    P3.12	P3.11    P3.10    P3.09    P3.08
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//______________________________________________________________________________
	;//Ports  :  P3.07    P3.06    P3.05    P3.04	P3.03    P3.02    P3.01    P3.00
	;//Signal :  I   	 I  	  I 	   I		I   	 I  	  I 	   I
	;//Binary :  0   	 0  	  0 	   0		0   	 0  	  0 	   0
	;//******************************************************************************

; Timer0 setup
T0IR			EQU		0xE0004000
T0TCR			EQU		0xE0004004
T0TC			EQU		0xE0004008
T0PR			EQU		0xE000400C
T0PC			EQU		0xE0004010
T0MCR			EQU		0xE0004014
T0MR0			EQU		0xE0004018
T0MR1			EQU		0xE000401C
T0MR2			EQU		0xE0004020
T0MR3			EQU		0xE0004024
T0CCR			EQU		0xE0004028
T0CR0			EQU		0xE000402C
T0CR1			EQU		0xE0004030
T0CR2			EQU		0xE0004034
T0CR3			EQU		0xE0004038
T0EMR			EQU		0xE000403C

T0PR_val		EQU		0x00000010
T0PC_val		EQU		0x0000000A
T0TCR_val_rst 	EQU		0x00000003	   ;enable and reset T0
T0TCR_val_start EQU		0x00000001	   ;enable T0
T0MCR_val		EQU		0x00000007	   ;interrupt on match MR0 and reset counter
T0MR0_val_15	EQU		0x0D693A40	   ;match value after 15s
T0MR0_val_5		EQU		0x047868C0	   ;match value after 5s
T0IR_M0_rst_val	EQU		0x00000001

; VIC Interrupt controller setup
VICIRQStatus	EQU		0xFFFFF000
VICFIQStatus	EQU		0xFFFFF004
VICRawIntr		EQU		0xFFFFF008
VICIntSelect	EQU		0xFFFFF00C
VICIntEnable	EQU		0xFFFFF010
VICIntEnClr		EQU		0xFFFFF014
VICSoftInt		EQU		0xFFFFF018
VICSoftIntClear	EQU		0xFFFFF01C
VICProtection	EQU		0xFFFFF020
VICVectAddr		EQU		0xFFFFF030
VICDefVectAddr	EQU		0xFFFFF034
VICVectAddr0	EQU		0xFFFFF100
VICVectAddr1	EQU		0xFFFFF104
VICVectAddr2	EQU		0xFFFFF108
VICVectAddr3	EQU		0xFFFFF10C
VICVectAddr4	EQU		0xFFFFF110
VICVectAddr5	EQU		0xFFFFF114
VICVectAddr6	EQU		0xFFFFF118
VICVectAddr7	EQU		0xFFFFF11C
VICVectAddr8	EQU		0xFFFFF120
VICVectAddr9	EQU		0xFFFFF124
VICVectAddr10	EQU		0xFFFFF128
VICVectAddr11	EQU		0xFFFFF12C
VICVectAddr12	EQU		0xFFFFF130
VICVectAddr13	EQU		0xFFFFF134
VICVectAddr14	EQU		0xFFFFF138
VICVectAddr15	EQU		0xFFFFF13C
VIVVectCntl0	EQU		0xFFFFF200
VIVVectCntl1	EQU		0xFFFFF204
VIVVectCntl2	EQU		0xFFFFF208
VIVVectCntl3	EQU		0xFFFFF20C
VIVVectCntl4	EQU		0xFFFFF210
VIVVectCntl5	EQU		0xFFFFF214
VIVVectCntl6	EQU		0xFFFFF218
VIVVectCntl7	EQU		0xFFFFF21C
VIVVectCntl8	EQU		0xFFFFF220
VIVVectCntl9	EQU		0xFFFFF224
VIVVectCntl10	EQU		0xFFFFF228
VIVVectCntl11	EQU		0xFFFFF22C
VIVVectCntl12	EQU		0xFFFFF230
VIVVectCntl13	EQU		0xFFFFF234
VIVVectCntl14	EQU		0xFFFFF238
VIVVectCntl15	EQU		0xFFFFF23C

VICIntEnable_val	EQU		0x00020010    ; enabling T0 and EINT3
VICIntSelect_val	EQU		0x00000010    ;	T0=FIQ, EINT3=IRQ
			
; Setup EINT3
EXTINT			EQU		0xE01FC140
EXTWAKE			EQU		0xE01FC144 
EXTMODE			EQU		0xE01FC148
EXTPOLAR		EQU		0xE01FC14C 

EXTINT_rst_val	EQU		0x00000008
EXTWAKE_val		EQU		0x00000008
EXTMODE_val		EQU		0x00000008
EXTPOLAR_val	EQU		0x00000008 

; Setup power control
PCON			EQU		0xE01FC0C0
PCON_val		EQU		0x00000001

				 PRESERVE8

; Area Definition and Entry Point
;  Startup Code must be linked first at Address at which it expects to run.

;###########################################################################################
; code portion including 
;     + vectors definition
;	  + peripheral core wake-up sequences
;	  + main program code
;	  + exception/interrupt handling routine
;##########################################################################################

; area RESET di tipo codice e in sola lettura: la scrittura fa scatenare 
; un'eccezione
                AREA    RESET, CODE, READONLY
                ENTRY

;###########################################################################################
; Exception vectors
;##########################################################################################

;  Mapped to Address 0.
;  Absolute addressing mode must be used.
;  Dummy Handlers are implemented as infinite loops which can be modified.

Vectors         LDR     PC, Reset_Addr         
                LDR     PC, Undef_Addr
                LDR     PC, SWI_Addr
                LDR     PC, PAbt_Addr
                LDR     PC, DAbt_Addr
                NOP                            ; Reserved Vector 
                LDR     PC, IRQ_Addr
                ;LDR     PC, [PC, #-0x0FF0]     ; Vector from VicVectAddr
                LDR     PC, FIQ_Addr

Reset_Addr      DCD     Reset_Handler
Undef_Addr      DCD     Undef_Handler
SWI_Addr        DCD     SWI_Handler
PAbt_Addr       DCD     PAbt_Handler
DAbt_Addr       DCD     DAbt_Handler
                DCD     0                      ; Reserved Address 
IRQ_Addr        DCD     IRQ_Handler
FIQ_Addr        DCD     FIQ_Handler

Undef_Handler   B       Undef_Handler
SWI_Handler     B       SWI_Handler
PAbt_Handler    B       PAbt_Handler
DAbt_Handler    B       DAbt_Handler

; Reset Handler          
Reset_Handler

;###########################################################################################
; peripheral core wake-up code
;###########################################################################################
   				
; Setup PLL (Phase locked loop) -> riceve una frequenza dall'esterno 
; (cristallino) e opera delle trasformazioni per ottenere la frequenza
; che ci interessa
                
                ; operazione di load che carica all'interno di R0 l'indirizzo
                ; di base del PLL
                LDR     R0, =PLL_BASE
                MOV     R1, #0xAA
                MOV     R2, #0x55

; Configure and Enable PLL
                MOV     R3, #PLLCFG_Val
                STR     R3, [R0, #PLLCFG_OFS] 
                MOV     R3, #PLLCON_PLLE
                ; operazioni di store di dati nel PLL
                STR     R3, [R0, #PLLCON_OFS]
                STR     R1, [R0, #PLLFEED_OFS]
                STR     R2, [R0, #PLLFEED_OFS]

; Wait until PLL Locked
PLL_Loop        LDR     R3, [R0, #PLLSTAT_OFS]
                ANDS    R3, R3, #PLLSTAT_PLOCK
                BEQ     PLL_Loop

; Switch to PLL Clock							 
                MOV     R3, #(PLLCON_PLLE:OR:PLLCON_PLLC)
                STR     R3, [R0, #PLLCON_OFS]
                STR     R1, [R0, #PLLFEED_OFS]
                STR     R2, [R0, #PLLFEED_OFS]

; Setup MAM				
                LDR     R0, =MAM_BASE
                MOV     R1, #MAMTIM_Val
                STR     R1, [R0, #MAMTIM_OFS] 
                MOV     R1, #MAMCR_Val
                STR     R1, [R0, #MAMCR_OFS] 
               
; Setup REMAP
                LDR     R0, =MEMMAP
                MOV     R1, #1
                STR     R1, [R0]

;  Setup PINSELx -> programmazione dei pin							   
                LDR     R5, =PINSEL0
                LDR     R6, =PINSEL0_val
				STR     R6, [R5]				
				LDR     R5, =PINSEL1
                LDR     R6, =PINSEL1_val
				STR     R6, [R5]				
				LDR     R5, =PINSEL2
                LDR     R6, =PINSEL2_val
				STR     R6, [R5]

; Setup IOxDIR		   
                LDR     R0, =IO0DIR
                LDR     R1, =IO0DIR_val
				STR     R1, [R0]
				
				LDR     R0, =IO1DIR
                LDR     R1, =IO1DIR_val
				STR     R1, [R0]
				
				LDR     R0, =IO2DIR
                LDR     R1, =IO2DIR_val
				STR     R1, [R0]

				LDR     R0, =IO3DIR
                LDR     R1, =IO3DIR_val
				STR     R1, [R0]

; Setup EXINT
				LDR		R0, =EXTWAKE
				LDR		R1, =EXTWAKE_val
				STR		R1, [R0]
				LDR		R0, =EXTMODE
				LDR		R1, =EXTMODE_val
				STR		R1, [R0]
				LDR		R0, =EXTPOLAR
				LDR		R1, =EXTPOLAR_val
				STR		R1, [R0]

; Initialize VIC -> vector interrupt controller: controllore degli 
; interrupt
				LDR		R0, =VICIntEnable
				LDR		R1, =VICIntEnable_val
				STR		R1, [R0]

				LDR		R0, =VICIntSelect
				LDR		R1, =VICIntSelect_val
				STR		R1, [R0]

;  Setup Stack for each mode
                LDR     R0, =Stack_Top

;  Enter Undefined Instruction Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_UND:OR:I_Bit:OR:F_Bit
                MOV     SP, R0
                SUB     R0, R0, #UND_Stack_Size

;  Enter Abort Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_ABT:OR:I_Bit:OR:F_Bit
                MOV     SP, R0
                SUB     R0, R0, #ABT_Stack_Size

;  Enter FIQ Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_FIQ:OR:I_Bit:OR:F_Bit
                MOV     SP, R0
                SUB     R0, R0, #FIQ_Stack_Size

;  Enter IRQ Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_IRQ:OR:I_Bit:OR:F_Bit
                MOV     SP, R0
                SUB     R0, R0, #IRQ_Stack_Size

;  Enter Supervisor Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_SVC:OR:I_Bit:OR:F_Bit
                MOV     SP, R0
                SUB     R0, R0, #SVC_Stack_Size

;  Enter User Mode and set its Stack Pointer
                MSR     CPSR_c, #Mode_USR
                MOV     SP, R0
                SUB     SL, SP, #USR_Stack_Size				
;###########################################################################################
; startup code, nel main si entra in modo utente
;###########################################################################################
main				
				LDR 	LR, =main   ; the operator = is used to copy the address of a label 			
				
				; stato iniziale: rosso per le auto e verde per i pedoni
				LDR		R0,	=IO0CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1,	[R0]
				
				LDR		R0,	=IO0SET
				LDR		R1, =IO0_red
				STR		R1,	[R0]

				LDR		R0,	=IO2CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1,	[R0]
				
				LDR		R0,	=IO2SET
				LDR		R1, =IO2_gre
				STR		R1,	[R0]
		
				; Contenuto R3:
				; stato 0: richiesta del pedone soddisfatta subito e semaforo pedoni verde
				; 	(STATO INIZIALE)
				; stato 1: richiesta del pedone soddisfatta subito e semaforo pedoni giallo
				; stato 2: in attesa di richieste e semaforo pedoni rosso
				; stato 3: richiesta del pedone soddisfatta dopo 10s
				; stato 4: richiesta del pedone soddisfatta dopo 5s
				AND		R3,	#0x00000000
				; salvo lo stato nello stack
				STMFD	SP!, {R3}

				;  Setup Timer0		 to count 15s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_15
				STR     R1, [R0]

up_power_down_state
                ; sequenza di basso consumo di potenza (simile a HLT nell'8086)		
				LDR		R0, =PCON
				LDR		R1, =PCON_val
				STR     R1, [R0]
				B		up_power_down_state		

				;startup end
				BX      LR				; this statement is never reached

;###########################################################################################
; IRQ handler: associated to the external interrupt EXTINT3    (button s9)
;###########################################################################################

IRQ_Handler     
                ; salvo il contesto
				SUBS 	LR, LR, #4
				; salvataggio di tutti i registri
				STMFD 	SP!,{R0-R12, LR}
				
				; identifichiamo l'interrupt
				LDR		R0, =VICIRQStatus
				LDR		R1, [R0] 			;read which IRQ is attivated
				; maschero gli altri interrupt
				ANDS	R1, #0x00020000
				BNE		EINT3_lab
				B		IRQ_end

			    ; se  arrivato un interrupt				
EINT3_lab
				; carico lo stato
				MOV		R0, SP
				SUB		R0, R0, #56
				LDR 	R3, [R0]	
				; se siamo nello stato 0
				CMP		R3, #0
				; resettiamo l'interrupt e torniamo al main	
				BEQ		Set_timer

				; se siamo nello stato 1
				CMP		R3, #1
				; resettiamo l'interrupt, scendiamo allo stato 0 e torniamo al main	
				BEQ		STAT_0

				; se siamo nello stato 2
				CMP		R3, #2
				; resettiamo l'interrupt, andiamo allo stato 3, settiamo il timer e torniamo al main	
				BEQ		NUOVA_REQ
				
				; se siamo nello stato 3 o 4
				CMP		R3, #2
				; resettiamo l'interrupt e torniamo al main	
				BGT		IRQ_rst


NUOVA_REQ
				; saliamo allo stato 3
				ADD		R3, R3, #1
				STR		R3, [R0]

				;  Setup Timer0		 to count 5s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_5
				STR     R1, [R0]

				; resettiamo l'interrupt
				B		IRQ_rst

STAT_0
				; supponendo che mi trovi nello stato 1
				; scendo allo stato 0
				SUB		R3, R3, #1
				STR		R3, [R0]
Set_timer
				;  Setup Timer0		 to count 15s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_15
				STR     R1, [R0]
IRQ_rst
                ; resetto l'interrupt esterno appena scatenato
				LDR		R0, =EXTINT
				LDR		R1, =EXTINT_rst_val
				STR		R1, [R0] 			;clear EXINT3 interrupt request
				

IRQ_end			; return to main
				; ripristino il contenuto dei registri				
				LDMFD 	SP!,{R0-R12, PC}^

;###########################################################################################
; FIQ handler: associated to the Timer0 interrupt on match 
;###########################################################################################

FIQ_Handler
                ; salvo l'indirizzo di ritorno     
				SUBS 	LR, LR, #4
				; salvo il contesto (i registri)
				STMFD 	SP!,{R0-R12, LR}    ;register dump into stack
				; carico l'indirizzo dell'interrupt controller
				LDR		R0, =VICFIQStatus
				; carico il valore dell'interrupt scatenato
				LDR		R1, [R0] 			;read which IRQ is activated
				; lo maschero con il valore che mi interessa
				ANDS	R1, #0x00000010
				; se  l'interrupt che mi interessa eseguo il codice
				; all'etichetta T0_lab
				BNE		T0_lab
				; altrimenti termino
				B		T0_end
								
T0_lab	
				; carico lo stato in R3
				MOV		R0, SP
				SUB		R0, R0, #56
				LDR 	R3, [R0]	
				; se lo stato  0 sono passati 15 secondi e passo allo stato 1
				CMP		R3, #0
				BEQ		Pedoni_giallo
				; se lo stato  1 sono passati 5 secondi e passo allo stato 2
				CMP		R3, #1
				BEQ		Pedoni_rosso
				; se lo stato  3 sono passati 5 secondi e passo allo stato 4
				CMP		R3, #3
				BEQ		Auto_giallo
				; se lo stato  4 sono passati 5 secondi e passo allo stato 0
				CMP		R3, #4
				BEQ		Pedoni_verde

				B		T0_end

Pedoni_giallo
				; passo allo stato 1
				ADD		R3, R3, #1
				STR		R3, [R0]
                ; carico l'indirizzo del pin dei led (semaforo pedoni)			
                LDR		R0, =IO2CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO2SET
				LDR		R1, =IO2_yel
				; il semaforo dei pedoni diventa giallo
				STR     R1, [R0]
				; carico l'indirizzo del timer 0
				LDR		R0, =T0IR
				; resetto il timer 0 e metto il valore da raggiungere a 5 secondi
				LDR		R1, =T0IR_M0_rst_val
				STR		R1, [R0] 			;clear Timer0 interrupt request
				;  Setup Timer0		 to count 5s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_5
				STR     R1, [R0]

				B		T0_end

Pedoni_rosso
				; passo allo stato 2
				ADD		R3, R3, #1
				STR		R3, [R0]
                ; carico l'indirizzo del pin dei led (semaforo pedoni)			
                LDR		R0, =IO2CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO2SET
				LDR		R1, =IO2_red
				; il semaforo dei pedoni diventa giallo
				STR     R1, [R0]
				; carico l'indirizzo del pin dei led (semaforo auto)			
                LDR		R0, =IO0CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO0SET
				LDR		R1, =IO0_gre
				; il semaforo delle auto diventa verde
				STR     R1, [R0]
				; carico l'indirizzo del timer 0
				LDR		R0, =T0IR
				; resetto il timer 0
				LDR		R1, =T0IR_M0_rst_val
				STR		R1, [R0] 			;clear Timer0 interrupt request
				
				B 		T0_end

Auto_giallo
				; passo allo stato 4
				ADD		R3, R3, #1
				STR		R3, [R0]
                ; carico l'indirizzo del pin dei led (semaforo auto)			
                LDR		R0, =IO0CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO0SET
				LDR		R1, =IO0_yel
				; il semaforo delle auto diventa giallo
				STR     R1, [R0]
				; carico l'indirizzo del timer 0
				LDR		R0, =T0IR
				; resetto il timer 0 e metto il valore da raggiungere a 5 secondi
				LDR		R1, =T0IR_M0_rst_val
				STR		R1, [R0] 			;clear Timer0 interrupt request
				;  Setup Timer0		 to count 5s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_5
				STR     R1, [R0]
				
				B		T0_end

Pedoni_verde
				; passo allo stato 0
				SUB		R3, R3, #4
				STR		R3, [R0]
                ; carico l'indirizzo del pin dei led (semaforo auto)			
                LDR		R0, =IO0CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO0SET
				LDR		R1, =IO0_red
				; il semaforo delle auto diventa rosso
				STR     R1, [R0]
				; carico l'indirizzo del pin dei led (semaforo pedoni)			
                LDR		R0, =IO2CLR
				LDR		R1, =0xFFFFFFFF
				STR		R1, [R0]
				LDR		R0, =IO2SET
				LDR		R1, =IO2_gre
				; il semaforo dei pedoni diventa giallo
				STR     R1, [R0]
				; carico l'indirizzo del timer 0
				LDR		R0, =T0IR
				; resetto il timer 0 e metto il valore da raggiungere a 15 secondi
				LDR		R1, =T0IR_M0_rst_val
				STR		R1, [R0] 			;clear Timer0 interrupt request
				;  Setup Timer0		 to count 5s
				;LDR     R0, =T0PR 			; prescale register for large count
                ;LDR     R1, =T0PR_val
				;STR     R1, [R0]
				;LDR     R0, =T0PC
                ;LDR     R1, =T0PC_val
				;STR     R1, [R0]			; not used in this example
				; registro di abilitazione					                
				LDR     R0, =T0TCR
                LDR     R1, =T0TCR_val_rst
				STR     R1, [R0]
				; disabilito il reset				
				LDR     R1, =T0TCR_val_start
				STR     R1, [R0]
				; impostazione del tipo di funzione:
				; in questo caso parte un interrupt al termine del conteggio e
				; il timer si blocca				
				LDR     R0, =T0MCR
                LDR     R1, =T0MCR_val
				STR     R1, [R0]
				; match register: il registro che contiene il valore da
				; contare; nel momento in cui questo viene raggiunto si
				; scatena un interrupt e si resetta il contatore								
				LDR     R0, =T0MR0
                LDR     R1, =T0MR0_val_15
				STR     R1, [R0]

T0_end			; return to main
                ; carico il contesto				
				LDMFD 	SP!,{R0-R12, PC}^

				END
