/*In a house there are 9 players and the Great Brother sitting in a confessional room.
Each player may ask to enter the confessional whenever he wants (at random intervals (max 25 seconds).
The request is accepted if:
 − The confessional is free
 − The Great Brother is not tired (max 5 seconds)
If the Great Brother does not accept the player in 10 seconds, the player goes away.
When a player enter the confessional, he thinks (max 5 seconds) and then he selects two other players
When a player has been selected more than 4 times, he is eliminated.*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include "semaphore.h"

#define N_PLAYERS 9

int id_concorrente;
int sem_confessionale[2];
int sem_generale[2];
int fd[2]; //in questa pipe il concorrente dice al GF chi ha nominato

void concorrente(int);
void grande_fratello();
void eliminato(int);
void vincitore(int);
int check_vincitore(int *);

int main()
{
    pid_t concorrenti[N_PLAYERS];    
    int i, sono_figlio = 0;
    
    if(sem_create(sem_generale) < 0){
	printf("Errore nella creazione del semaforo generale!\n");
	exit(EXIT_FAILURE);
    }
    if(sem_create(sem_confessionale) < 0){
	printf("Errore nella creazione del semaforo del confessionale!\n");
	exit(EXIT_FAILURE);
    }
    if(pipe(fd) < 0){
	printf("Errore nella creazione della pipe!\n");
	exit(EXIT_FAILURE);
    }
    
    set_handler(SIGUSR1, eliminato);
    set_handler(SIGUSR2, vincitore);
    
    for(i = 0; i < N_PLAYERS; i++)
    {	
	pid_t temp = fork();
	if(temp == 0)
	{
	    sono_figlio = 1;
	    concorrente(i);
	    break;
	}
	else
	    concorrenti[i] = temp;
    }
    
    if(!sono_figlio)
	grande_fratello(concorrenti);
	
    return EXIT_SUCCESS;
}

void eliminato(int n)
{
    printf("****** Concorrente %d: sono stato eliminato!\n", id_concorrente);
    exit(EXIT_SUCCESS);
}

void vincitore(int n)
{
    printf("****** Concorrente %d: sono il vincitore!\n", id_concorrente);
    exit(EXIT_SUCCESS);
}

void concorrente(int me)
{
    srand(getpid());
    int nomination1, nomination2;    
    char buffer[2];
    id_concorrente = me;
    
    close(fd[0]);
    
    while(1)
    {
	sem_post(sem_generale);
	
	if (sem_trywait(sem_confessionale, 10) == 0){
	    sem_wait(sem_generale);//annullo la mia richiesta di accesso al confessionale
	    printf("Concorrente %d ha aspetto troppo per nominare. Ritorna dopo.\n", me);
	    sleep(rand()%25 + 1);
	    continue;
	}
	
	printf("Concorrente %d pensa a chi nominare.\n", me);
	
	sleep(rand()%5 + 1);
	do
	{
	    nomination1 = rand()%9;
	    nomination2 = rand()%9;
	}while(nomination1 == me || nomination2 == me);
	
	buffer[0] = '0' + nomination1;
	buffer[1] = '0' + nomination2;
	write(fd[1], buffer, 2);
	printf("NOMINATION: %d e %d.\n", nomination1, nomination2);
	
	sleep(rand()%25 + 1);
    }
}

void grande_fratello(pid_t * concorrenti)
{
    int voti[N_PLAYERS];
    int nominato1, nominato2;
    char buffer[2];
    int i, ret;
    
    close(fd[1]);
    srand(getpid());
    
    for(i = 0; i < N_PLAYERS; i++)
	voti[i] = 0;
    while(1)
    {
	printf("Il Grande Fratello si riposa.\n");
	sleep(rand()%5 + 1);
	printf("Il Grande Fratello ha finito di riposare.\n");
	
	if(sem_trywait(sem_generale, 0) == 0)
	    continue;
	
	//Qualcuno vuole entrare nel confessionale, lo faccio entrare
	sem_post(sem_confessionale);
	
	//Aspetto la votazione
	read(fd[0], buffer, 2);
	nominato1 = buffer[0] - '0';
	nominato2 = buffer[1] - '0';
	voti[nominato1]++;
	voti[nominato2]++;
	
	for(i = 0; i < N_PLAYERS; i++)
	    if(voti[i] >= 4)
		kill(concorrenti[i], SIGUSR1);
	    
	ret = check_vincitore(voti);
	if(ret == N_PLAYERS)
	{
	    printf("Nessun vincitore!\n");
	    break;
	}else if(ret == N_PLAYERS-1)
	{
	    //C'e' stato un vincitore
	    for(i = 0; i < N_PLAYERS; i++)
	    {
		if(voti[i] < 4){
		    kill(concorrenti[i], SIGUSR2);
		    break;
		}
	    }
	    break;
	}
    }
}

int check_vincitore(int * voti)
{
    int i, count = 0;
    for(i = 0; i < N_PLAYERS; i++)
	if(voti[i] >= 4)
	    count++;
    return count;
}
