#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include "semaphore.h"

int sem_create(int fd[2])
{
    if(pipe(fd))
	return -1;
    return 1;
}

int sem_init(int pipe[2], int val)
{
    int i;
    for(i = 0; i < val; i++)
	if(sem_post(pipe) < 0)
	    return -1;
    return 1;
}

int sem_trywait(int pipe[2], int timeout)
{
    fd_set fds;
    int n;
    struct timeval tv;
    char buffer[2];
    
    FD_ZERO(&fds);
    FD_SET(pipe[0], &fds);
    
    if(timeout < 0)
	n = select(pipe[0]+1, &fds, NULL, NULL, NULL);
    else{
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	n = select(pipe[0]+1, &fds, NULL, NULL, &tv);
    }
    
    if (n == 0) return 0;
    if (n == -1) return -1;
    if (read(pipe[0], &buffer, 1) != 1)
	return -1;
    return 1;
}

int sem_wait(int pipe[2])
{
    return sem_trywait(pipe, -1);
}

int sem_post(int pipe[2])
{    
    if (write(pipe[1], "X", 1) != 1)
	return -1;
    return 1;
}

int set_handler (int signo, SigFunc handler)
{
    /* Utilizziamo sa_handler, non sa_sigaction, pertanto
    non utilizziamo il flag SA_SIGINFO.
    Il segnale passato come argomento viene mascherato
    di default, a meno che non si utilizzi il flag
    SA_NODEFER o SA_NOMASK in sa_flags */
    
    struct sigaction act;
    memset(&act, 0, sizeof(act));
    act.sa_handler = handler;
    //act.sa_flags = SA_NODEFER; non ha effetto
    if(sigaction(signo, &act, NULL) == -1)
	return -1;
    return 1;
}
