/*Scrivere un programma C che realizzi il prodotto di due matrici utilizzando Pthread.
Siano [r1, c1] e [r2, cs] le dimensioni delle due matrici da moltiplicare (con c1 = r2).
Il thread principale crea un numeri di thread uguale a r1*c2 ognuno dei quali effettua
un singolo prodotto riga-colonna.
Al termine dell'esecuzione di tutti i thread, il thread principale si occupa di visualizzare
la matrice risultato*/

/* Variante: i thread creati dal thread principale a loro volta creano altri thread ciascuno
dei quali calcola il singolo elemento della matrice risultato. Questi si sincronizzano con i
thread creati dal main che calcolano un prodotto riga per colonna, questi si sincronizzano
col main per il calcolo complessivo della matrice prodotto */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include <assert.h>
#include <unistd.h>

#define ROW1 2
#define COL1 5
#define ROW2 5
#define COL2 3
#define MAXRAND 12

typedef struct
{
    int ** matr1;
    int ** matr2;
    int ** matr_ris;
    int row;
    int col;
    sem_t * sem;
    pthread_mutex_t ** mutexes_matr1;
    pthread_mutex_t ** mutexes_matr2;
}parameter;

typedef struct
{
    int ** matr1;
    int ** matr2;
    int row;
    int col;
    int n;
    int * ris;
    sem_t * sem;
    pthread_mutex_t ** mutexes_matr1;
    pthread_mutex_t ** mutexes_matr2;
}parameter2;

void print_matr(int **, int, int);
void * thread_function(void *);
void * worker_thread(void *);

int main()
{
    int numthreads, i, j, ** matr1, ** matr2, ** matr_ris;
    pthread_t * threads;
    pthread_mutex_t ** mutexes_matr1, ** mutexes_matr2;
    sem_t sem;
    parameter * par;
    
    srand(time(0)*getpid());
    assert(COL1 == ROW2);
    
    sem_init(&sem, 0, 0);
    matr1 = (int **) malloc(ROW1 * sizeof(int *));
    for(i = 0; i < ROW1; i++)
	matr1[i] = (int *) malloc(COL1 * sizeof(int));
    matr2 = (int **) malloc(ROW2 * sizeof(int *));
    for(i = 0; i < ROW2; i++)
	matr2[i] = (int *) malloc(COL2 * sizeof(int));
    matr_ris = (int **) malloc(ROW1 * sizeof(int *));
    for(i = 0; i < ROW1; i++)
	matr_ris[i] = (int *) malloc(COL2 * sizeof(int));

    for(i = 0; i < ROW1; i++)
	for(j = 0; j < COL1; j++)
	    matr1[i][j] = rand()%MAXRAND + 1;
    for(i = 0; i < ROW2; i++)
	for(j = 0; j < COL2; j++)
	    matr2[i][j] = rand()%MAXRAND + 1;
    for(i = 0; i < ROW1; i++)
	for(j = 0; j < COL2; j++)
	    matr_ris[i][j] = 0;
	
    printf("\nMatrice 1:\n");
    print_matr(matr1, ROW1, COL1);
    printf("\n");
    printf("Matrice 2:\n");
    print_matr(matr2, ROW2, COL2);
    printf("\n");
    
    numthreads = ROW1 * COL2;
    threads = (pthread_t *) malloc(numthreads * sizeof(pthread_t));
    mutexes_matr1 = (pthread_mutex_t **) malloc(ROW1 * sizeof(pthread_mutex_t *));
    for(i = 0; i < ROW1; i++)
	mutexes_matr1[i] = (pthread_mutex_t *)malloc(COL1 * sizeof(pthread_mutex_t));
    for(i = 0; i < ROW1; i++)
	for(j = 0; j < COL1; j++)
	    pthread_mutex_init(&mutexes_matr1[i][j], NULL);
    mutexes_matr2 = (pthread_mutex_t **) malloc(ROW2 * sizeof(pthread_mutex_t *));
    for(i = 0; i < ROW2; i++)
	mutexes_matr2[i] = (pthread_mutex_t *)malloc(COL2* sizeof(pthread_mutex_t));
    for(i = 0; i < ROW2; i++)
	for(j = 0; j < COL2; j++)
	    pthread_mutex_init(&mutexes_matr2[i][j], NULL);
    
    par = (parameter *) malloc(ROW1 * COL2 * sizeof(parameter));
    for(i = 0; i < ROW1; i++)
    {
	for(j = 0; j < COL2; j++)
	{
	    par[(i*COL2)+j].matr1 = matr1;
	    par[(i*COL2)+j].matr2 = matr2;
	    par[(i*COL2)+j].matr_ris = matr_ris;
	    par[(i*COL2)+j].row = i;
	    par[(i*COL2)+j].col = j;
	    par[(i*COL2)+j].sem = &sem;
	    par[(i*COL2)+j].mutexes_matr1 = mutexes_matr1;
	    par[(i*COL2)+j].mutexes_matr2 = mutexes_matr2;
	}
    }

    for(i = 0; i < numthreads; i++)
	pthread_create(&threads[i], NULL, thread_function, (void *)&par[i]);
    
    for(i = 0; i < numthreads; i++)
	sem_wait(&sem);
    
    printf("Matrice prodotto:\n");
    print_matr(matr_ris, ROW1, COL2);
    printf("\n");
    
    for(i = 0; i < ROW1; i++)
	free(matr1[i]);
    free(matr1);
    for(i = 0; i < ROW2; i++)
	free(matr2[i]);
    free(matr2);
    for(i = 0; i < ROW1; i++)
	free(matr_ris[i]);
    free(matr_ris);
    
    free(threads);
    for(i = 0; i < ROW1; i++)
	free(mutexes_matr1[i]);
    free(mutexes_matr1);
    for(i = 0; i < ROW2; i++)
	free(mutexes_matr2[i]);
    free(mutexes_matr2);
    
    return EXIT_SUCCESS;
}

void print_matr (int ** m, int r, int c)
{
    int i, j;
    for(i = 0; i < r; i++){
	for(j = 0; j < c; j++)
	    printf ("%4d", m[i][j]);
	printf("\n");
    }
}

void * thread_function (void * arg)
{
    parameter * par = (parameter *) arg;
    parameter2 * par2;
    pthread_t * threads;
    int ** matr1 = par->matr1;
    int ** matr2 = par->matr2;
    int ** matr_ris = par->matr_ris;
    int r = par->row;
    int c = par->col;
    pthread_mutex_t ** mutexes_matr1 = par->mutexes_matr1;
    pthread_mutex_t ** mutexes_matr2 = par->mutexes_matr2;
    sem_t sem;
    int val_prodotto = 0, i, * ris;
    
    sem_init(&sem, 0, 0);
    
    threads = (pthread_t *) malloc(COL1 * sizeof(pthread_t));
    ris = (int *) malloc(COL1 * sizeof(int));
    for(i = 0; i < COL1; i++)
	ris[i] = 0;
    par2 = (parameter2 *) malloc(COL1 * sizeof(parameter2));
    for(i = 0; i < COL1; i++)
    {
	par2[i].matr1 = matr1;
	par2[i].matr2 = matr2;
	par2[i].row = r;
	par2[i].col = c;
	par2[i].n = i;
	par2[i].ris = &ris[i];
	par2[i].mutexes_matr1 = mutexes_matr1;
	par2[i].mutexes_matr2 = mutexes_matr2;
	par2[i].sem = &sem;
    }

    for(i = 0; i < COL1; i++)
	pthread_create(&threads[i], NULL, worker_thread, (void *)&par2[i]);
    for(i = 0; i < COL1; i++)
	sem_wait(&sem);
    
    for(i = 0; i < COL1; i++)
	val_prodotto += *(par2[i].ris);
    matr_ris[r][c] = val_prodotto;

    sem_post(par->sem);
    
    return NULL;
}

void * worker_thread (void * arg)
{
    parameter2 * par = (parameter2 *) arg;
    
    pthread_mutex_lock(&par->mutexes_matr1[par->row][par->n]);
    pthread_mutex_lock(&par->mutexes_matr2[par->n][par->col]);
    
    *(par->ris) = par->matr1[par->row][par->n] * par->matr2[par->n][par->col];
    
    pthread_mutex_unlock(&par->mutexes_matr2[par->n][par->col]);
    pthread_mutex_unlock(&par->mutexes_matr1[par->row][par->n]);
    
    sem_post(par->sem);
    
    return NULL;
}