/* Due file contengono stringhe di lunghezza pari a 9 caratteri (piu' il carattere di "a capo") su ciascuna riga).
Si scriva un programma concorrente in cui il main-padre sgancia due processi figli che leggono ciascuno uno dei due file.
Le righe lette vengono inviate al padre tramite pipe. Il padre visualizza a video le stringhe ricevute, con l'indicazione del processo
figlio che le ha lette e a lui inviate. Al termine della lettura del proprio file ciascun figlio termina e invia al padre la stringa "end".
Il padre termina una volta ricevuta la stringa "end" di entrambi i figli*/

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

char * file_1 = "file_1";
char * file_2 = "file_2";

void sem_wait(int *);
void sem_post(int *);
void figlio(char *, int *);
void padre(int *, pid_t, int *, pid_t);

int main()
{
    pid_t figlio1, figlio2;
    int pipe1[2], pipe2[2];
    
    if(pipe(pipe1)){
	printf("Errore nella creazione della pipe1!\n");
	exit(EXIT_FAILURE);
    }
    if(pipe(pipe2)){
	printf("Errore nella creazione della pipe2!\n");
	exit(EXIT_FAILURE);
    }
    
    if((figlio1 = fork()) == 0)
    {
	//sono il primo figlio1
	figlio(file_1, pipe1);
    }
    else
    {
	if((figlio2 = fork()) == 0)
	{
	    //sono il secodo figlio
	    figlio(file_2, pipe2);
	}
	else
	    padre(pipe1, figlio1, pipe2, figlio2);
    }
    
    waitpid(-1, NULL, 0);
    waitpid(-1, NULL, 0);
    
    return EXIT_SUCCESS;
}

void sem_wait(int * pipe)
{
    char buffer[2];
    if (read(pipe[0], &buffer, 1) != 1) {
	printf ("Error on sem_wait\n");
	exit (EXIT_FAILURE);
    }
}

void sem_post(int * pipe)
{    
    if (write(pipe[1], "X", 1) != 1) {
	printf ("Error on sem_post\n");
	exit (EXIT_FAILURE);
    }
}

void padre(int * pipe1, pid_t figlio1, int * pipe2, pid_t figlio2)
{
    int ret, max, end_rcvd = 0;
    char buffer[20];
    fd_set fds;
    
    close(pipe1[1]);
    close(pipe2[1]);
    
    FD_ZERO(&fds);
    FD_SET(pipe1[0], &fds);
    FD_SET(pipe2[0], &fds);
    max = (pipe1[0] > pipe2[0] ? pipe1[0] : pipe2[0]) + 1;
    
    while(1)
    {
	ret = select(max, &fds, NULL, NULL, NULL);
	if(FD_ISSET(pipe1[0], &fds))
	{
	    if (read(pipe1[0], buffer, 10) <= 0) {
		printf ("Errore lettura da pipe1!\n");
		exit (EXIT_FAILURE);
	    }
	    if(!strcmp(buffer, "endendend")){
		end_rcvd++;
		break;
	    }
	    else
		buffer[9] = '\0';
	    printf("[%d] Ricevuto %s da figlio1\n", getpid(), buffer);
	    FD_SET(pipe2[0], &fds); //perchè?
	}
	else if(FD_ISSET(pipe2[0], &fds))
	{
	    if (read(pipe2[0], buffer, 10) <= 0) {
		printf ("Errore lettura da pipe2!\n");
		exit (EXIT_FAILURE);
	    }
	    if(!strcmp(buffer, "endendend")){
		end_rcvd++;
		break;
	    }
	    else
		buffer[9] = '\0';
	    printf("[%d] Ricevuto %s da figlio2\n", getpid(), buffer);
	    FD_SET(pipe1[0], &fds); //perchè?
	}
	if(end_rcvd == 2)
	    break;
    }
    
    FD_CLR(pipe1[0], &fds);
    FD_CLR(pipe2[0], &fds);
}

void figlio(char * file, int * pipe)
{
    FILE * f;
    char buffer[20];
    int i;
    
    srand(time(0)+getpid());
    if((f = fopen(file, "r")) == NULL){
	printf("Errore apertura file %s\n", file);
	exit(EXIT_FAILURE);
    }
    while(fgets(buffer, 11, f) != NULL)
    {
	buffer[9] = '\0';
	printf("[%d] Read from %s: %s\n", getpid(), file, buffer);
	if (write(pipe[1], buffer, 10) <= 0) {
	    printf ("Error on write\n");
	    exit (EXIT_FAILURE);
	}
	sleep(rand()%5+1);
    }
    
    if (write(pipe[1], "endendend", 9) <= 0) {
	printf ("Error on write\n");
	exit (EXIT_FAILURE);
    }
    printf("[%d] ha finito!\n", getpid());    
}