/*
Realizzare in OS/161 un programma concorrente in linguaggio C che, utilizzando i thread,
manipoli delle stringhe di caratteri, con le seguenti specifiche.
Il thread principale attiva due thread ulteriori (denominati sender e receiver)
e si mette in attesa che entrambi terminino.
Il thread sender procede con costrutto iterativo all'interno del quale:
 - legge una stringa da standard input (tastiera)
 - segnala al receiver che la stringa è stata letta e aspetta che questo la manipoli
 - quando il receiver ha manipolato la stringa e sveglia il sender, quest'ultimo visualizza la stringa su standard output (video)
Il thread receiver include un costrutto iterativo all'interno del quale:
 - attende la stringa dal sender
 - trasforma i caratteri della stringa in maiuscoli (qualora siano alfabetici)
 - sveglia il sender e si rimette in attesa della stringa successiva.

I cicli di lettura/visualizzazione e di conversione in maiuscolo terminano non appena viene introdotta la stringa “end”.
Il thread principale visualizzi il numero di stringhe processate prima di terminare definitivamente.*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <string.h>

#define BUF_LEN 1024

char buffer[BUF_LEN];
int tot;

typedef void SigFunc (int);
void set_handler (int, SigFunc);
void * receiver(void *);
void * sender(void *);
void StrToUpper(char *);
sem_t can_read, can_write;

int main()
{
    pthread_t sender_th, receiver_th;
    sem_init(&can_read, 0, 0);
    sem_init(&can_write, 0, 0);
    tot = 0;
    
    set_handler(SIGINT, SIG_IGN); /* Ignora Ctrl-C */
    set_handler(SIGTSTP, SIG_IGN);/* Ignora Ctrl-Z */
    pthread_create(&sender_th, NULL, sender, NULL);
    pthread_create(&receiver_th, NULL, receiver, NULL);
    
    pthread_join(sender_th, NULL);
    pthread_join(receiver_th, NULL);
    
    printf("Totale messaggi: %d\n", tot);
    
    return EXIT_SUCCESS;
}

void * sender (void * arg)
{
    while(1)
    {
	printf("Inserisci la stringa: ");
	fgets(buffer, BUF_LEN, stdin);
	sem_post(&can_read);
	
	if(!strcmp(buffer, "end\n"))
	    break;
	
	sem_wait(&can_write);
	
	printf("Stringa: %s\n", buffer);
    }
    return NULL;
}

void * receiver (void * arg)
{    
    while(1)
    {
	sem_wait(&can_read);
	
	if(!strcmp(buffer, "end\n"))
	    break;
	tot++;
	StrToUpper(buffer);
	
	sem_post(&can_write);
    }
    
    return NULL;
}

void StrToUpper(char * s)
{
    int i;
    char * alfa = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    
    for(i = 0; s[i] != '\n'; i++)
	if(s[i] >= 'a' && s[i] <= 'z')
	    s[i] = alfa[s[i] - 'a'];
    s[i] = '\0';
}


void set_handler (int signo, SigFunc handler)
{
    /* Utilizziamo sa_handler, non sa_sigaction, pertanto
    non utilizziamo il flag SA_SIGINFO.
    Il segnale passato come argomento viene mascherato
    di default, a meno che non si utilizzi il flag
    SA_NODEFER o SA_NOMASK in sa_flags */
    
    struct sigaction act;
    memset(&act, 0, sizeof(act));
    act.sa_handler = handler;
    //act.sa_flags = SA_NODEFER; non ha effetto
    if(sigaction(signo, &act, NULL) == -1)
	printf("ERROR on sigaction\n");
}

