/* Si scriva in OS/161 un programma che sia in grado di ordinare un vettore di n+1 elementi utilizzando n thread identici.
Ogni thread è responsabile di due elementi adiacenti (il thread 0 si occupa degli elementi 0 e 1,
il thread 1 degli elementi 1 e 2, ..., il thread n-1 degli elementi n-1 e n).
Ciascun thread confronta i due elementi di cui si occupa e li scambia se non si trovano nell'ordine corretto.
Una volta terminato il proprio lavoro tutti i thread si sincronizzano con un thread specifico che termina l'intero processo
se tutti gli elementi sono ordinati correttamente oppure esegue nuovamente tutti i thread per proseguire nella fase di ordinamento.*/

#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdlib.h>
#include <time.h>

#define VETT_SIZE 20
#define MAXRAND  100
#define TRUE  1
#define FALSE 0

int FINITO;

int * vett;
pthread_mutex_t * mutexes;
sem_t barriera1, barriera2, barriera3;

void init_vett(int *);
int is_sorted(int *);
void * sorter(void *);
void print_vett(int *);

int main()
{
    int i, * pi;
    pthread_t * threads;
    
    srand(time(0));
    
    FINITO = FALSE;
    sem_init (&barriera1, 0, VETT_SIZE-1);
    sem_init (&barriera2, 0, 0);
    sem_init (&barriera3, 0, 0);
    
    vett = (int *) malloc(VETT_SIZE * sizeof(int));
    init_vett(vett);
    print_vett(vett);
    threads = (pthread_t *) malloc((VETT_SIZE - 1) * sizeof(pthread_t));
    mutexes = (pthread_mutex_t *) malloc((VETT_SIZE) * sizeof(pthread_mutex_t));
    
    for(i = 0; i < VETT_SIZE; i++)
	pthread_mutex_init (&mutexes[i], NULL);
    for(i = 0; i < VETT_SIZE - 1; i++){    
	pi = (int *) malloc (sizeof (int));
	*pi = i;
	pthread_create(&threads[i], NULL, sorter, pi);
    }
	
    while(1)
    {
	for(i = 0; i < VETT_SIZE-1; i++)
	    sem_wait(&barriera2);

	if(is_sorted(vett)){
	    FINITO = TRUE;
	    for(i = 0; i < VETT_SIZE-1; i++)
		sem_post(&barriera1);
	    for(i = 0; i < VETT_SIZE-1; i++)
		sem_post(&barriera3);
	    break;
	}
	print_vett(vett);
	for(i = 0; i < VETT_SIZE-1; i++)
	    sem_post(&barriera1);
	for(i = 0; i < VETT_SIZE-1; i++)
	    sem_post(&barriera3);
	
    }
    print_vett(vett);
    return 0;
}

void * sorter(void * arg)
{
    int temp, index = *((int *) arg);;
    while(1)
    {
	sem_wait(&barriera1);
	if(FINITO) break;
	
	pthread_mutex_lock(&mutexes[index]);
	pthread_mutex_lock(&mutexes[index+1]);	
	
	if(vett[index] > vett[index+1]){
	    temp = vett[index];
	    vett[index] = vett[index+1];
	    vett[index+1] = temp;
	}
	sem_post(&barriera2);
	pthread_mutex_unlock(&mutexes[index]);	
	pthread_mutex_unlock(&mutexes[index+1]);
	sem_wait(&barriera3);
    }
    
    return NULL;
}

void init_vett(int * v)
{
    int i;
    for(i = 0; i < VETT_SIZE; i++)
	v[i] = rand()%MAXRAND + 1;
}

int is_sorted(int * v)
{
    int i;
    for(i = 0; i < VETT_SIZE - 1; i++)
	if(v[i] > v[i+1])
	    return FALSE;
    return TRUE;
}

void print_vett(int * v)
{
    int i;
    for(i = 0; i < VETT_SIZE; i++)
	printf("%d ", v[i]);
    printf("\n");
}