#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#define TRENI 10
// N.B. Se TRENI > 10 (numero su 2 cifre) occorre modificare R/W da pipe
#define USCITE 4

int pipe_in[2], pipe_out[2], semafori[TRENI][2];

void sem_init (int *s) { pipe (s); }
void sem_signal (int* s) { write (s[1], "X", 1); }
void sem_wait (int* s) { char b; read (s[0], &b, 1); }

int
casuale (
  float max
  )
{
  struct timeb tb;

  ftime (&tb);
  srand (tb.millitm);
  return (int) (max*rand()/(RAND_MAX+1.0));
}

void
treno (
  // t = id del treno (0 - TRENI-1)
  int t
  )
{
  char buf[2];
  int c;

  /* Treno in viaggio prima di arrivare in stazione */
  c = casuale (15.0);
  printf ("Treno %d: attesa di %d.\n", t, c);
  sleep (c);
 
  /* Treno "t" in ingresso con destinazione binario "c" */ 
  buf[0] = (char) t;
  c = casuale ((float) USCITE);
  printf ("Treno %d: partenza per %d\n", t, c);
  buf[1] = (char) casuale(c);
  write (pipe_in[1], buf, 2);
  sem_wait (semafori[t]);
 
  /* Treno entrato e poi in attesa di carico/scarico passeggeri */
  printf ("Treno %d: entrato in stazione per linea di uscita %d.\n", t, buf[1]);
  c=casuale(5.0);
  printf ("Treno %d: attesa di %d.\n", t, c);
  sleep (c);

  /* Treno in uscita */
  write (pipe_out[1], &buf[1], 1);
  printf ("Treno %d: uscito.\n", t);
  exit (0);
}

void
stazione (
  )
{
  fd_set fds;
  int r, i, max;
  char buf[2];
  char flag;

  /* Il vettore seguente contiene un elemento per ciascun treno. */
  /* Ogni elemento vale 0 se il treno corrispondente non  in */
  /* attesa di entrare, 1 altrimenti. */
  char waiting [TRENI];

  /* Il vettore seguente contiene il numero della linea di uscita */
  /* che il treno dovr percorrere. */
  int destinazioni [TRENI];

  /* Il vettore seguente contiene un elemento per ciascuna linea */
  /* di uscita. Ogni elemento vale 0 se la linea  libera. */
  char lineeUscita [USCITE];

  for (i=0; i<TRENI; i++) {
    waiting [i] = 0;
    destinazioni [i] = -1;
  }

  for (i=0; i<USCITE; i++) {
    lineeUscita [i] = 0;
  }

  FD_ZERO (&fds);
  FD_SET (pipe_in[0], &fds);
  FD_SET (pipe_out[0], &fds);
  max = (pipe_in[0]>pipe_out[0]?pipe_in[0]:pipe_out[0])+1;

  while (1) {
    r=select (max, &fds, NULL, NULL, NULL);
    if (FD_ISSET(pipe_in[0], &fds)) {
      FD_SET (pipe_out[0], &fds);
      /* buf[0]  lid del treno */
      /* buf[1]  la linea di uscita che vuole percorrere */
      read (pipe_in[0], &buf, 2);
      if (lineeUscita[(int) buf[1]] == 0) {
        sem_signal (semafori[(int) buf[0]]);
        lineeUscita [(int) buf[1]] = 1;
      } else {
        destinazioni [(int) buf[0]] = buf[1];
      }
    } else {
      FD_SET (pipe_in[0], &fds);
      /* Leggo la linea che si  liberata */
      read (pipe_out[0], &buf, 1);
      /* Cerco un treno che vuole andare su questa linea */
      flag = 0;
      for (i=0; i<TRENI; i++) {
        if (flag==0 && destinazioni[i]==buf[0]) {
          destinazioni[i] = -1;
          sem_signal (semafori[i]);
          flag = 1;
        }
      }
      if (flag == 0) {
        lineeUscita[(int) buf[0]] = 0;
      }
    }
  }
}

int
main (
  )
{
  int i;

  pipe (pipe_in);
  pipe (pipe_out);
  for (i=0; i<TRENI; i++) {
    sem_init (semafori[i]);
  }
  for (i=0; i<TRENI; i++) {
    if (fork()==0) {
      /* Figlio */
      treno (i);
    }
  }
  printf ("Parte il processo stazione.\n");
  stazione ();
  return 0;
}
