/*Si implementi, in linguaggio C, in ambiente Unix/Linux, un programma in grado di duplicare una albero di direttori
(i.e., un programma che realizzi il comando unix “cp –R”).
Il programma riceve sulla riga di comando due stringhe rappresentanti il pathname completo (path+nome) di due direttori.
Il programma replica il contenuto del primo direttorio (esistente) nel secondo (inesistente).
Per duplicazione si intende una copia del primo direttorio nel secondo, inclusi tutti i sottodirettori e tutti i file,
ottenendo al termine del procedimento la stessa struttura gerarchica ad albero e gli stessi contenuti.*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>

#define MAX_LEN 1024

void RecursiveCopy(const char *, const char *);
void CopyFile(const char *, const char *);

int main(int argc, char ** argv)
{
    char temp[MAX_LEN];
    if(argc != 3){
	printf("Usage: %s <source_dir> <destination_dir>\n", argv[0]);
	exit(EXIT_FAILURE);
    }    
    if(mkdir(argv[2], S_IRWXU|S_IRWXG|S_IRWXO)){
	sprintf(temp, "Error on creating '%s'", argv[2]);
	perror(temp);
	exit(EXIT_FAILURE);
    }
    
    RecursiveCopy(argv[1], argv[2]);
    return EXIT_SUCCESS;
}

void RecursiveCopy(const char * source, const char * dest)
{
    DIR * dir;
    struct dirent * dir_struct;
    char buffer[MAX_LEN], buffer2[MAX_LEN];
    
    dir = opendir(source);
    if(dir == NULL)
    {
	sprintf(buffer, "Error on opening '%s'", source);
	perror(buffer);
	exit(EXIT_FAILURE);
    }
    
    while((dir_struct = readdir(dir)) != NULL)
    {
	if(dir_struct->d_type == DT_DIR && !(!strcmp(dir_struct->d_name, ".") || !strcmp(dir_struct->d_name, "..")))
	{   
	    sprintf(buffer, "%s/%s", dest, dir_struct->d_name);
	    if(mkdir(buffer, S_IRWXU|S_IRWXG|S_IRWXO))
	    {
		sprintf(buffer, "Error on creating '%s'", buffer);
		perror(buffer);
		exit(EXIT_FAILURE);
	    }
	    bzero(buffer, MAX_LEN);
	    sprintf(buffer, "%s/%s", source, dir_struct->d_name);
	    sprintf(buffer2, "%s/%s", dest, dir_struct->d_name);
	    RecursiveCopy(buffer, buffer2);
	}
	else if(dir_struct->d_type == DT_REG)
	{
	    sprintf(buffer, "%s/%s", source, dir_struct->d_name);
	    sprintf(buffer2, "%s/%s", dest, dir_struct->d_name); 
	    CopyFile(buffer, buffer2);
	}
    }    
}

void CopyFile(const char * source, const char * dest)
{
    FILE * f1, * f2;
    char buffer[MAX_LEN];
    char ch;
	
    if((f1 = fopen(source, "r")) == NULL){
	sprintf(buffer, "Error on reading '%s'", source);
	perror(buffer);
	exit(EXIT_FAILURE);
    }
    if((f2 = fopen(dest, "w+")) == NULL){
	sprintf(buffer, "Error on creating '%s'", dest);
	perror(buffer);
	exit(EXIT_FAILURE);
    }
    
    while(!feof(f1))
    {
	ch = fgetc(f1);
	fputc(ch, f2);
    }   
   
    fclose(f1);
    fclose(f2);
	
	/* int nR, nW, fdR, fdW;
	char buf[MAX_LEN];
	
	fdR = open(source, O_RDONLY);
	fdW = open(dest, O_WRONLY|O_CREAT);
	
	while((nR = read(fdR, buf, BUFFSIZE)) > 0)
	{
		nW = write(fdW, buf, nR);
		if(nR != nW)
			fprintf(stderr, "Errore in scrittura!\n");
	}
	close(fdR);
	close(fdW);
	*/
}