/*Sviluppare un server TCP (in ascolto sulla porta specificata come primo parametro sulla riga di
comando) che accetti richieste di trasferimento file da client ed invii il file richiesto.
Sviluppare un client che si colleghi ad un server TCP (all’indirizzo e porta specificati come primo e
secondo parametro sulla riga di comando) per richiedere dei file e memorizzarli localmente. I nomi
dei file da richiedere vengono forniti su standard input, uno per riga. Ogni file richiesto deve essere
salvato localmente e deve essere fornito su standard output un messaggio circa l’avvenuto
trasferimento, con nome e dimensione del file.
Per richiedere un file il client invia al server i tre caratteri ASCII “GET” seguiti dal nome del file
(max 64 caratteri) e da CR LF (senza alcun spazio):

GET nomefile CR LF

Il server risponde inviando:

+OK CR LF

seguito dal numero di byte (intero su 32 bit in formato di rete) e quindi dai byte del file in oggetto.
Il client può richiedere più file inviando più comandi GET. Quando intende terminare la
comunicazione invia:

QUIT CR LF

e chiude il canale.
In caso di errore (es. comando illegale, file inesistente) il server risponde sempre con

–ERR CR LF

e quindi chiude il canale col client.
Provare a collegare il proprio client col server sviluppato da un altro gruppo e viceversa.
Mentre è in corso un collegamento provare ad attivare un secondo client verso il medesimo server.
Provare ad attivare sul medesimo nodo una seconda istanza del server sulla medesima porta.
Provare a collegare il client ad un indirizzo non raggiungibile.
Provare a collegare il client ad un indirizzo esistente ma ad una porta su cui il server non è in
ascolto.
Provare a disattivare il server (battendo ^C nella sua finestra) mentre un client è collegato.
*/

//CLIENT

#include "sockwrap.h"
#include "errlib.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#define MAX_LEN 1024

char * prog_name;
void Receive_File(char *, int, int);

int main(int argc, char *argv[])
{
    int sock, file_size = 0;
    short port;
    char buffer[MAX_LEN];
    struct sockaddr_in servaddr;
    struct hostent * server;
    
    prog_name = argv[0];
    
    if (argc != 3)
	err_quit ("usage: %s <address> <port>\n", prog_name);
    port = atoi(argv[2]);
    
    sock = Socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    err_msg("(%s) socket created", prog_name);
    
    server = Gethostbyname(argv[1]);	
    
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET; //AF_INET, AF_INET6, AF_UNSPEC
    servaddr.sin_port = htons(port);
    memcpy((char *)server->h_addr, (char *)&servaddr.sin_addr.s_addr, server->h_length);
    
    Connect(sock, (struct sockaddr*) &servaddr, sizeof(servaddr));
     
    fgets(buffer, MAX_LEN, stdin);

    Send(sock, buffer, strlen(buffer)+1, 0);
    
    memset(buffer, 0, MAX_LEN);
    
    Read(sock, buffer, MAX_LEN);
    printf("%s", buffer);
    
    if(!strcmp(buffer, "-ERR\n")) return 0;
    
    memset(buffer, 0, MAX_LEN);
    
    printf("Ricezione dimensione file\n");
    Readn(sock, &file_size, sizeof(int));
    file_size = ntohl(file_size);
    
    printf("Ricezione file (%d bytes)\n", file_size);
    Receive_File("ciao", file_size, sock);
    
    Close(sock);
    
    return 0;
}

void Receive_File(char * name, int size, int sock)
{
    char buffer[MAX_LEN];
    int nRead = 0, nWrite = 0, fdW;
    
    fdW = open (name, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    while(size > 0)
    {
	nRead = read(sock, buffer, MAX_LEN);
	nWrite = write(fdW, buffer, nRead);
	if(nRead != nWrite)
		err_msg("(%s) ERROR - Write Error (read %d, write %d)\n", prog_name, nRead, nWrite);
	size -= nRead;
    }
    close (fdW);
}