/*
 * Lab 2 - Esercizio 2.3
 * client_test_tcp
 * Author: Carmelo Clemenza
 * Date: March 2011
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>	// O_WRONLY | O_CREAT | O_TRUNC
#include <sys/stat.h>	// S_IRUSR | S_IWUSR
#include <math.h>	// log10()

#include "errlib.h"
#include "sockwrap.h"

#ifdef TRACE
#define trace(x) x
#else
#define trace(x)
#endif

#define QUIT		0
#define OK	 	1
#define ERR		2
#define FILE_NAME	"Rec\0"
#define MSG_OK 	 	"+OK\n"
#define MSG_ERR  	"-ERR\n"
#define MSG_QUIT 	"QUIT\n"
#define MAX_BUFF_IN 	64
#define MAX_BUFF_OUT 	1024

#define FILE_STR "prova.txt"

char *prog_name;

int is_ok(char *str, int len){
	if( strncmp(str, MSG_OK, strlen(MSG_OK)) == 0){
		trace( err_msg("(%s) INFO - received string is '%s'", prog_name, MSG_OK) );
		return OK;
	}
	
	if( strncmp(str, MSG_ERR, strlen(MSG_ERR)) == 0){
		trace( err_msg("(%s) INFO - received string is '%s'", prog_name, MSG_ERR) );
		return ERR;
	}
	
	if( strncmp(str, MSG_QUIT, strlen(MSG_QUIT)) == 0 ){
		trace( err_msg("(%s) INFO - received string is '%s'", prog_name, MSG_QUIT) );
		return QUIT;
	}
	
	return -1;
}

void save_file(int sock_fd, u_int32_t size, char* r_buf, u_int32_t r_buf_size, int index){
	int nR=0, nW=0, fdW;		// for write the file
	char *f_name;
	
	/* compute chars to be used for writing index */
	double n = log10(index);
	n = floor(n + 1);
	
	f_name = (char*) malloc(strlen(FILE_NAME) + 1 + n + 1);	// +2 to make room for '-' and '\0'
	sprintf(f_name, "%s-%d", FILE_NAME, index);
	
	fdW = open (f_name, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
	if(fdW == (-1)) {
		err_sys("File Open Error (W=%d)\n", fdW);
	}
	trace( err_msg("(%s) INFO - open file to write '%s'", prog_name, f_name);
	
	while(size > 0){
		nR = read(sock_fd, r_buf, r_buf_size);
		nW = write(fdW, r_buf, nR);
		if(nR != nW)
			err_msg("(%s) ERROR - Write Error (read %d, write %d)\n", prog_name, nR, nW);
		size = size - nR;
	}
	if (nW < 0)
		err_msg("(%s) ERROR - Write Error\n", prog_name);
	
	trace( err_msg("(%s) INFO - writing file '%s' finished", prog_name, f_name) );
	trace( err_msg("(%s) INFO - close file to write '%s'", prog_name, f_name) );
	close (fdW);
	free(f_name);
}

int send_command(int sock_fd){
	char cmd[MAX_BUFF_IN + 4]; 	//+4 to make room for "GET" and "\n"
	char r_buffer[MAX_BUFF_OUT];
	u_int32_t size;
	ssize_t n, cmd_size;
	int i = 0;
	
	while(1){
		printf("(%s) Insert command to send to server: ", prog_name);
		scanf("%s", cmd);
		
		trace( err_msg("(%s) INFO - new command to send to server '%s'", prog_name, cmd) );
		n = strlen(cmd);		
		if(n > MAX_BUFF_IN+3){
			err_msg("(%s) ERROR - max length reached (%d)", prog_name, n);
			continue;
		}
		cmd[n++]='\n';		
		cmd_size = n;
		
		Writen(sock_fd, cmd, n);
		trace( err_msg("(%s) INFO - new command sended to server", prog_name) );
		trace( err_msg("(%s) INFO - wait for answer", prog_name) );
		
		n = MyReadline(sock_fd, r_buffer, MAX_BUFF_OUT);
		
		switch( is_ok(r_buffer, n) ){
			case OK:
				err_msg("(%s) INFO - server reply '+OK\\n'", prog_name);
				break;
			case ERR:
				err_msg("(%s) INFO - server reply '-ERR\\n'", prog_name);
				continue;
			case QUIT:
				err_msg("(%s) INFO - server reply 'QUIT\\n'", prog_name);
				return i;
			default:
				err_msg("(%s) ERROR - wrong reply from server", prog_name);
				return i;		
		}
		
		/* Receive file size */
		n = Readn(sock_fd, &size, sizeof(u_int32_t));
		if(n < sizeof(u_int32_t)){
			err_msg("(%s) ERROR - read only %uB while I waited %uB", prog_name, 
					n, sizeof(u_int32_t));
			return i;
		}
		size = ntohl(size);
		err_msg("(%s) INFO - file size is %uB", prog_name, size);
		
		/* Receive file content */
		save_file(sock_fd, size, r_buffer, MAX_BUFF_OUT, i+1);
		
		i++;
		trace( err_msg("(%s) INFO - file %i received with success", prog_name, i) );
	}
	
	return i;
}

int main (int argc, char *argv[]) {
	int socket_fd, err=0, n_file;
	short port;
	struct sockaddr_in serv_addr;

	/* for errlib to know the program name */
	prog_name = argv[0];

	/* check arguments */
	if (argc!=3)
		err_quit("usage: %s <address> <port>\n where <address> can be 127.0.0.1 or other", prog_name);
	
	port=atoi(argv[2]);
	if(port <= 0) {
		err_sys("(%s) ERROR - atoi() failed", prog_name);
	}

	/* create socket */
	socket_fd = Socket(AF_INET, SOCK_STREAM, 0);

	/* specify address to connect to */
	memset(&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);
	Inet_pton (AF_INET, argv[1], (void*) &serv_addr.sin_addr.s_addr);

	trace( err_msg("(%s) INFO - socket created",prog_name) );
	trace( err_msg ("(%s) INFO - waiting for connections ...", prog_name) );

	Connect(socket_fd, (SA*) &serv_addr, sizeof(serv_addr));
	trace( err_msg("(%s) INFO - new connection to server %s:%u", 
		prog_name, inet_ntoa(serv_addr.sin_addr), ntohs(serv_addr.sin_port)) );

	n_file = send_command(socket_fd);
	trace( err_msg("(%s) INFO - %d file(s) transfer with success", prog_name, n_file) );
	
	Close(socket_fd);
	trace( err_msg ("(%s) - connection closed by %s", prog_name, (err==0)?"server":"client") );

	return 0;
}

