/*Sviluppare un server TCP (in ascolto sulla porta specificata come primo parametro sulla riga di
comando) che accetti richieste di trasferimento file da client ed invii il file richiesto.
Sviluppare un client che si colleghi ad un server TCP (all’indirizzo e porta specificati come primo e
secondo parametro sulla riga di comando) per richiedere dei file e memorizzarli localmente. I nomi
dei file da richiedere vengono forniti su standard input, uno per riga. Ogni file richiesto deve essere
salvato localmente e deve essere fornito su standard output un messaggio circa l’avvenuto
trasferimento, con nome e dimensione del file.
Per richiedere un file il client invia al server i tre caratteri ASCII “GET” seguiti dal nome del file
(max 64 caratteri) e da CR LF (senza alcun spazio):

GET nomefile CR LF

Il server risponde inviando:

+OK CR LF

seguito dal numero di byte (intero su 32 bit in formato di rete) e quindi dai byte del file in oggetto.
Il client può richiedere più file inviando più comandi GET. Quando intende terminare la
comunicazione invia:

QUIT CR LF

e chiude il canale.
In caso di errore (es. comando illegale, file inesistente) il server risponde sempre con

–ERR CR LF

e quindi chiude il canale col client.
Provare a collegare il proprio client col server sviluppato da un altro gruppo e viceversa.
Mentre è in corso un collegamento provare ad attivare un secondo client verso il medesimo server.
Provare ad attivare sul medesimo nodo una seconda istanza del server sulla medesima porta.
Provare a collegare il client ad un indirizzo non raggiungibile.
Provare a collegare il client ad un indirizzo esistente ma ad una porta su cui il server non è in
ascolto.
Provare a disattivare il server (battendo ^C nella sua finestra) mentre un client è collegato.
*/

//SERVER

#include "sockwrap.h"
#include "errlib.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#define MAX_LEN 1024
#define VERO 1
#define FALSO 0

char * prog_name;
void get_file_name(char *, char *);
int get_file_size(char *);
int check_command(char *);
void Send_File(char *, int, int);

int main(int argc, char *argv[])
{
    int file_size, listenfd, connfd, reuseaddr = 1;
    short port;
    struct sockaddr_in servaddr, cliaddr;
    socklen_t cliaddrlen = sizeof(cliaddr);
    char buffer[MAX_LEN];
    char file_name[MAX_LEN];
            
    prog_name = argv[0];
    
    if (argc != 2)
	err_quit ("usage: %s <port>\n", prog_name);
    port = atoi(argv[1]);
    
    listenfd = Socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(int)) == -1)
	err_quit("(%s) socket options", prog_name);
    err_msg("(%s) socket created", prog_name);
    
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET; //AF_INET, AF_INET6, AF_UNSPEC
    servaddr.sin_port = htons(port);
    servaddr.sin_addr.s_addr = INADDR_ANY;
    
    Bind(listenfd, (SA*) &servaddr, sizeof(servaddr));
    
    Listen(listenfd, 1);
    err_msg("(%s) listening on %s:%u", prog_name, inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));
    
    while (1)
    {
	err_msg ("(%s) waiting for connections ...", prog_name);
	connfd = Accept (listenfd, (SA*) &cliaddr, &cliaddrlen);
	err_msg("(%s) - new connection from client %s:%u", prog_name, inet_ntoa(cliaddr.sin_addr), ntohs(cliaddr.sin_port));
	
	while(1)
	{
	    Read(connfd, buffer, MAX_LEN);
	    if(check_command(buffer) == FALSO){
		printf("%s: Comando non riconosciuto!\n", buffer);
		Write(connfd, "-ERR\n", 5);
		Close (connfd);
		break;
	    }
	    
	    if(!strcmp(buffer, "QUIT\n")){
		Close (connfd);
		break;
	    }
	    	    
	    //Il comando e' una GET di un file
	    get_file_name(file_name, buffer);
	    memset(buffer, 0, MAX_LEN);
	    printf("File Name: %s\n", file_name);
	    if((file_size = get_file_size(file_name)) < 0){
		printf("File inesistente!\n");
		Write(connfd, "-ERR\n", 5);
		Close (connfd);
		break;
	    }	        
	    Write(connfd, "+OK\n", 4);
	    
	    printf("Invio dimensione file\n");
	    file_size = htonl(file_size);
	    Writen(connfd, &file_size, sizeof(int));	    
	    file_size = ntohl(file_size);
	    printf("Invio file %s (%d bytes)\n", file_name, file_size);
	    Send_File(file_name, file_size, connfd);
	    printf("File inviato.\n");
	    break;
	}
    }
    return 0;
}

int check_command(char * command)
{    
    if((command[0] == 'G' && command[1] == 'E' && command[2] == 'T' && command[3] == ' ') ||	
       (command[0] == 'Q' && command[1] == 'U' && command[2] == 'I' && command[3] == 'T' && command[4] == '\n'))
	return VERO;
    return FALSO;
}

void get_file_name(char * to, char * from)
{
    int index = 0, i;
    memset(to, 0, MAX_LEN);
    for(i = 4; from[i] != '\n' && from[i] != '\0'; i++)
	to[index++] = from[i];
}

int get_file_size(char * file_name)
{
    FILE * f;
    int size;    
    if((f = fopen(file_name, "r")) == NULL)
	return -1;
    fseek(f, 0, SEEK_END);
    size = ftell(f);
    fclose(f);
    return size;
}

void Send_File(char * name, int size, int sock)
{
    /*FILE * f;
    char buffer[MAX_LEN];
    int read, LeftToSend = size;
    f = fopen(name, "r");
    
    while(LeftToSend > 0){
	read = fread(buffer, 1, MAX_LEN, f);
	Writen(sock, buffer, read);
	//printf("Scritti %d\n", read);
	LeftToSend -= read;
	//printf("LeftToSend: %d\n", LeftToSend);
    }
    fclose(f);*/
    int fdR, nR;
    char buffer[MAX_LEN];
    fdR = open(name, O_RDONLY);

    while ((nR = Read(fdR, buffer, MAX_LEN)) > 0)
		Writen(sock, buffer, nR);
    close(fdR);
}