/*
 * Lab 2 - Esercizio 2.3
 * server_test_tcp
 * Author: Carmelo Clemenza
 * Date: March 2011
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>	// O_RDONLY

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <string.h>
#include <time.h>

#include "errlib.h"
#include "sockwrap.h"

#define LISTENQ  15
#define MAX_BUFF_IN 64
#define MAX_BUFF_OUT 1024

#define MSG_OK 	 "+OK\n"
#define MSG_ERR  "-ERR\n"
#define MSG_QUIT "QUIT\n"
#define MSG_GET	 "GET"

#ifdef TRACE
#define trace(x) x
#else
#define trace(x)
#endif

char *prog_name;


void send_file(int connfd, char* file){
	int nR=0, fdR;
	u_int32_t size;
	char buf[MAX_BUFF_OUT];

	fdR = open(file, O_RDONLY);
	if( fdR == (-1) ){
		err_ret("(%s) ------ File Open Error (R=%d)\n", prog_name, fdR);
		trace( err_msg("(%s) --- Reply to client '-ERR\\n'", prog_name) );
		Write(connfd, MSG_ERR, strlen(MSG_ERR));
		return;
	}
	
	/* compute file size */
	size = lseek(fdR, 0L, SEEK_END);
	if( size < 0 ){
		err_ret("(%s) ------ ERROR: fseek() failed\n", prog_name);
		trace( err_msg("(%s) --- Reply to client '-ERR\\n'", prog_name) );
		Write(connfd, MSG_ERR, strlen(MSG_ERR));
		return;
	}
	if(( lseek(fdR, 0L, SEEK_SET) ) < 0 ){
		err_ret("(%s) ------ ERROR: fseek() failed\n", prog_name);
		trace( err_msg("(%s) --- Reply to client '-ERR\\n'", prog_name) );
		Write(connfd, MSG_ERR, strlen(MSG_ERR));
		return;
	}
	
	/* file is ok */
	trace( err_msg("(%s) --- Reply to client '+OK\\n'", prog_name) );
	Write(connfd, MSG_OK, strlen(MSG_OK));
	
	/* send file size to client */
	trace( err_msg("(%s) --- Reply to client file size %dB", prog_name, size) );
	size = htonl(size);
	Writen(connfd, &size, sizeof(u_int32_t));
	
	/* send file to client */
	
	trace( err_msg("(%s) --- Reply to client file content", prog_name) );
	while ((nR = Read(fdR, buf, MAX_BUFF_OUT)) > 0) {
		Writen(connfd, buf, nR);
	}
	
	close(fdR);
}

void receive_command(int connfd){
	char buf[MAX_BUFF_IN + 5]; 	/* +5 to make room for "GET" and "\n\0" */
	char file_name[MAX_BUFF_IN+1];	/* +1 to make room for \0 */
	int nread, i;

	while (1) {
		trace( err_msg("(%s) --- waiting for operands ...",prog_name) );
		nread = MyReadline(connfd, buf, MAX_BUFF_IN+4);
		if(nread == 0)
			return;

		/* append the string terminator after CR-LF */
		buf[nread]='\0';
		trace( err_msg("(%s) --- received string '%s'",prog_name, buf) );
		
		/* check command */		
		if( strncmp(buf, MSG_QUIT, strlen(MSG_QUIT)) == 0 ){
			trace( err_msg("(%s) INFO - client want to exit", prog_name) );
			Write (connfd, MSG_QUIT, strlen(MSG_QUIT));
			return;
		}

		if( strncmp(buf, MSG_GET, strlen(MSG_GET)) != 0){
			trace( err_msg("(%s) --- wrong or missing operands", prog_name) );
			Write (connfd, MSG_ERR, strlen(MSG_ERR));
			continue;
		}
		
		/* get file_name from buffer */
		nread = nread - 4;		// -4 for "GET" and '\0'
		strncpy(file_name, buf+3, nread);
		file_name[nread] = '\0';

		trace( err_msg("(%s) --- operands '%s'", prog_name, file_name) );
		
		send_file(connfd, file_name);
		
		trace( err_msg("(%s) --- result just sent back", prog_name) );
	}
}

int main (int argc, char *argv[]) {
	int listenfd, connfd, err=0;
	short port;
	struct sockaddr_in servaddr, cliaddr;
	socklen_t cliaddrlen = sizeof(cliaddr);

	/* for errlib to know the program name */
	prog_name = argv[0];

	/* check arguments */
	if(argc!=2)
		err_quit("usage: %s <port>\n", prog_name);
	port=atoi(argv[1]);
	if(port <= 0) {
		err_sys("(%s) error - atoi() failed", prog_name);
	}

	/* create socket */
	listenfd = Socket(AF_INET, SOCK_STREAM, 0);

	/* specify address to bind to */
	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	Bind(listenfd, (SA*) &servaddr, sizeof(servaddr));

	trace ( err_msg("(%s) socket created",prog_name) );
	trace ( err_msg("(%s) listening on %s:%u", prog_name, inet_ntoa(servaddr.sin_addr), 
						   ntohs(servaddr.sin_port)) );

	Listen(listenfd, LISTENQ);

	while (1) {
		trace( err_msg ("(%s) waiting for connections ...", prog_name) );

		connfd = Accept (listenfd, (SA*) &cliaddr, &cliaddrlen);
		trace( err_msg("(%s) - new connection from client %s:%u", 
			prog_name, inet_ntoa(cliaddr.sin_addr), ntohs(cliaddr.sin_port)) );

		receive_command(connfd);

		Close(connfd);
		trace( err_msg ("(%s) - connection closed by %s", prog_name, (err==0)?"client":"server") );
	}
	return 0;
}

