/*
	PROVINCE(COD_P, Province, Region)
	TIMEDIM(COD_T, Month, Quarter, FourMonth_Period, Year)
	CATEGORYDIM(COD_C, Category)
	RENTAL(COD_T, COD_C, COD_P, NetIncome, NumberOfRentedItems)
*/

-- Query 1
-- Calcolare per ogni coppia regione-trimestre l'incasso totale.
-- Calcolare, inoltre, l'incasso totale per ogni regione e l'incasso totale per ogni trimestre

SELECT Region, Quarter, SUM(NetIncome) AS TotIncome,
	   SUM(SUM(NetIncome)) OVER (PARTITION BY Region)  AS TotIncomePerRegion,
	   SUM(SUM(NetIncome)) OVER (PARTITION BY Quarter) AS TotIncomPerQuarter
FROM RENTAL R, PROVINCE P, TIMEDIM T
WHERE R.COD_P = P.COD_P AND R.COD_T = T.COD_T
GROUP BY Region, Quarter
ORDER BY Region, Quarter

-- Query 2
-- Calcolare l'incasso totale per ogni coppia [categoria oggetto affittato, mese].
-- Calcolare, inoltre, l'incasso totale per categoria di oggetto (indipendentemente
-- dal mese) e l'incasso  totale per ogni mese (indipendentemente dalla categoria di
-- oggetti affittati).

SELECT Category, Month, SUM(NetIncome) AS TotIncome,
	   SUM(SUM(NetIncome)) OVER (PARTITION BY Category) AS TotIncomePerCategory,
	   SUM(SUM(NetIncome)) OVER (PARTITION BY Month)    AS TotIncomePerMonth
FROM RENTAL R, TIMEDIM T, CATEGORY C
WHERE R.COD_T = T.COD_T AND R.COD_C = C.COD_C
GROUP BY Category, Month
ORDER BY Category, Month

-- Query 3
-- Calcolare il numero totale di oggetti affittati per ogni coppia [categoria ogetto affittato, mese].
-- Calcolare, inoltre, il numero totale di oggetti affittati per categoria di oggetto
-- (indipendentemente dal mese) e il numero totale di oggetti affittati per ogni mese
-- (indipendentemente dalla categoria di oggetti affittati)

SELECT Category, Month, SUM(NumberOfRentedItems) AS TotRented,
	   SUM(SUM(NumberOfRentedItems)) OVER (PARTITION BY Category) AS TotRentedPerCategory,
	   SUM(SUM(NumberOfRentedItems)) OVER (PARTITION BY Month)    AS TotRentedPerMonth
FROM RENTAL R, TIMEDIM T, CATEGORY C
WHERE R.COD_T = T.COD_T AND R.COD_C = C.COD_C
GROUP BY Category, Month
ORDER BY Category, Month
